#!/usr/bin/perl
#
#
# This program find peaks in a 2D fiber based spectral image
#f
#
#        ($yfit, $coeffs) = my_linfit1d $pdl_flux_masked,$pdl_model_tmp,$pdl_e;

use Statistics::OLS;
use Math::FFT;
use Math::Stat;
use Math::Spline qw(spline linsearch binsearch);
use Math::Derivative qw(Derivative2);

use Math::Approx;


use Astro::FITS::CFITSIO qw( :longnames :constants );
use PDL;

use PDL::Fit::Polynomial; 
use PDL::Filter::Linear;
use PGPLOT;  # Load PGPLOT module
use PDL::Fit::Gaussian;
use PDL::Image2D;
use  PDL::Fit::Linfit;



use PDL::Core;
use PDL::Basic;
use PDL::Exporter;
@ISA    = qw( PDL::Exporter );
use PDL::Options ':Func';
use PDL::Slatec; # For matinv()



srand(time() | $$);

$vel_light=299792.458;
$red_elines=0.0;

sub my_linfit1d {
   my $opthash = ref($_[-1]) eq "HASH" ? pop(@_) : {} ; 
   my %opt = parse( { Weights=>ones(1) }, $opthash ) ;
   barf "Usage: linfit1d incorrect args\n" if $#_<1 or $#_ > 3;
   my ($y, $fitfuncs, $wt) = @_;
   if ($#_ == 1) {
      ($y, $fitfuncs) = @_;
      $x = $y->xvals;
   }
   
#   my $wt = $opt{Weights};
   
   # Internally normalise data
   
   my $ymean = (abs($y)->sum)/($y->nelem);
   $ymean = 1 if $ymean == 0;
   my $y2 = $y / $ymean;
   
   # Do the fit
      
   my $M = $fitfuncs->xchg(0,1);
   my $C = $M->xchg(0,1) x ($M * $wt->dummy(0)) ;
   my $Y = $M->xchg(0,1) x ($y2->dummy(0) * $wt->dummy(0));

   # Fitted coefficients vector

   my   $a = matinv($C) x $Y;
   
   # Fitted data

   my $yfit = ($M x $a)->clump(2); # Remove first dim=1
   
   $yfit *= $ymean; # Un-normalise
   if (wantarray) {
      my $coeff = $a->clump(2);
      $coeff *= $ymean; # Un-normalise
      return ($yfit, $coeff);
   }
   else{
      return $yfit;
   }  
   
}
*my_linfit1d = \&my_linfit1d;

$F3D=$ENV{FIT3D_PATH}; $F3D=$ENV{FIT3D_PATH}; 
require("$F3D/scripts/my.pl");
#require("$F3D/scripts/my_fit_elines_rnd.pl");

$sec_ini=print_time();

#require "/work1/ssa/perl/MY/my.pl";
#require "/home/sanchez/sda1/perl/MY/my.pl";

$time1=get_time();
if ($#ARGV<5) {
    print "USE: auto_ssp.pl SPEC1.txt SSP_SFH.fits OUTFILE MASK_LIST CONFIG_FILE PLOT [min max] [wmin wmax] [redshift_elines_to_mask] [input_redshift delta_redshift min_redshift max_redshift] [input_sigma delta_sigma min_sigma max_sigma] [input_Av delta_Av min_Av max_Av] \n";
    print "CONFIG_FILE:\n";
    print "redshift delta_redshift min_redshift max_redshift\n";
    print "sigma delta_sigma min_sigma max_sigma\n";
    print "Av delta_Av min_Av max_Av [Same range for all]\n";
    print "N_SYSTEMS\n";
    print "(1) START_W END_W MASK_FILE CONFIG_FILE NPOLY MASK_FILE_POLY\n";
    print "...\n";
    print "(N) START_W END_W MASK_FILE CONFIG_FILE NPOLY MASK_FILE_POLY\n";
    print "MIN_DELTA_CHISQ MAX_NITER CUT_MEDIAN_FLUX\n";
    print "start_w_peak end_w_peak\n";
    print "wavelength_to_norm width_AA new_back_templates.fits\n";


    exit;
}

$n_mc=10;

$unc_file=$ARGV[0];
$clean_file="clean_".$ARGV[0];

$back_list=$ARGV[1];
$outfile=$ARGV[2];
$out_elines="elines_".$outfile;
$out_single="single_".$outfile;
$out_fit="fit_".$outfile;
$out_coeffs_file="coeffs_".$outfile;
$out_fit="output.".$outfile.".fits";
$out_ps=$outfile;

#######################################
# Clean previous results
#######################################
$call="rm -rf ".$outfile;
system($call);
$call="rm -rf ".$out_elines;
system($call);
$call="rm -rf ".$out_single;
system($call);
$call="rm -rf ".$out_fit;
system($call);
#$call="rm -rf ".$out_coeffs_file;
#system($call);

$D_SYS_VEL=100;



$mask_list=$ARGV[3];
$config_file=$ARGV[4];
$plot=$ARGV[5];

$ENV{PGPLOT_FOREGROUND} = "white";
$ENV{PGPLOT_BACKGROUND} = "black";

if ($plot==2) {
    $ENV{PGPLOT_FOREGROUND} = "black";
    $ENV{PGPLOT_BACKGROUND} = "white";
    $dev_plot=$outfile.".ps/CPS";
    $dev_plot_single="single_".$outfile.".ps/CPS";
} else {
    $dev_plot="/xs";
    $dev_plot_single="/xs";
}

$smooth=1;
$MIN_CHISQ=1e12;


$out_file="junk.junk";
$factor=1;
$box=1;


$def=0;
if ($#ARGV==7) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $def=1;
}

if ($#ARGV==9) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $def=2;
}

if ($#ARGV==10) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $elines_mask=$ARGV[10];
    $def=2;
}

$input_redshift=0;
if ($#ARGV==11) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $elines_mask=$ARGV[10];
    $input_redshift=$ARGV[11];
    $redshift=$input_redshift;
    $def=2;
}




#print "$#ARGV $d_redshift\n"; 
open(FH,"<$config_file");
$line=<FH>;
chop($line);
($redshift,$d_redshift,$min_redshift,$max_redshift,$DV,$RV,$DS,$RS,$MIN_W,$MAX_W)=split(" ",$line);


if ($#ARGV==14) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $elines_mask=$ARGV[10];
    $input_redshift=$ARGV[11];
    $input_d_redshift=$ARGV[12];
    $input_min_redshift=$ARGV[13];
    $input_max_redshift=$ARGV[14];
    $redshift=$input_redshift;
    $d_redshift=$input_d_redshift;
    $min_redshift=$input_min_redshift;
    $max_redshift=$input_max_redshift;
    $def=2;
}

$line=<FH>;
chop($line);
($sigma,$d_sigma,$min_sigma,$max_sigma)=split(" ",$line);

if ($#ARGV==18) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $elines_mask=$ARGV[10];
    $input_redshift=$ARGV[11];
    $input_d_redshift=$ARGV[12];
    $input_min_redshift=$ARGV[13];
    $input_max_redshift=$ARGV[14];
    $sigma=$ARGV[15];
    $d_sigma=$ARGV[16];
    $min_sigma=$ARGV[17];
    $max_sigma=$ARGV[18];
    $redshift=$input_redshift;
    $d_redshift=$input_d_redshift;
    $min_redshift=$input_min_redshift;
    $max_redshift=$input_max_redshift;
    $def=2;
}




$line=<FH>;
chop($line);
($Av_IN,$d_Av_IN,$min_Av,$max_Av)=split(" ",$line);
if ($#ARGV==22) {
    $min=$ARGV[6];
    $max=$ARGV[7];
    $min_wave=$ARGV[8];
    $max_wave=$ARGV[9];
    $elines_mask=$ARGV[10];
    $input_redshift=$ARGV[11];
    $input_d_redshift=$ARGV[12];
    $input_min_redshift=$ARGV[13];
    $input_max_redshift=$ARGV[14];
    $sigma=$ARGV[15];
    $d_sigma=$ARGV[16];
    $min_sigma=$ARGV[17];
    $max_sigma=$ARGV[18];
    $Av_IN=$ARGV[19];
    $d_Av_IN=$ARGV[20];
    $min_Av=$ARGV[21];
    $max_Av=$ARGV[22];
    $redshift=$input_redshift;
    $d_redshift=$input_d_redshift;
    $min_redshift=$input_min_redshift;
    $max_redshift=$input_max_redshift;
    $def=2;
}



#if ($input_redshift>0) {
#    $redshift=$input_redshift;
#    $d_redshift=$input_d_redshift;
#    $min_redshift=$input_min_redshift;
#    $max_redshift=$input_max_redshift;
#}



$REDSHIFT=$redshift;
$Av_ini=$Av_IN;

if ($d_redshift!=0) {
    $fit_redshift=1;
} else {
    $fit_redshift=0;
}

print "FIT_RED $fit_redshift $d_redshift $#ARGV\n"; #<stdin>;



$line=<FH>;
chop($line);
$ns=$line;
$start_w_min=1e12;
$end_w_max=-1e12;
for ($i=0;$i<$ns;$i++) {
    $line=<FH>;
    chop($line);
    ($start_w_e,$end_w_e,$mask_e,$config_e,$npoly_e,$mask_poly_e,$nmin_e,$nmax_e)=split(" ",$line);
    $start_w_E[$i]=$start_w_e;
    $end_w_E[$i]=$end_w_e;
    $mask_E[$i]=$mask_e;
    $config_E[$i]=$config_e;
#
# We read all the information
#
    my $n_line=0;
    open(CONF,"<$config_e");
    while($conf=<CONF>) {
	chop($conf);
	$config_line_E[$i][$n_line]=$conf;
	$n_line++;
    }
    close(CONF);
    $n_line_E[$i]=$n_line;


    $npoly_E[$i]=$npoly_e;
    $mask_poly_E[$i]=$mask_poly_e;
    $nmin_E[$i]=$nmin_e;
    $nmax_E[$i]=$nmax_e;
    if ($start_w_e<$start_w_min) {
	$start_w_min=$start_w_e;
    }
    if ($end_w_e>$end_w_max) {
	$end_w_max=$end_w_e;
    }
}
$line=<FH>;
chop($line);
($MIN_DELTA_CHISQ,$MAX_NITER,$CUT_MEDIAN_FLUX)=split(" ",$line);
$ABS_MIN=0.5*$CUT_MEDIAN_FLUX;
#print "$CUT_MEDIAN_FLUX\n"; exit;



$line=<FH>;
chop($line);
($start_w_peak,$end_w_peak)=split(" ",$line);
$line=<FH>;
chop($line);
($wave_norm,$w_wave_norm,$new_back_file)=split(" ",$line);
close(FH);


$pdl_flux_c_ini=rfits($back_list);
($n_c,$nf)=$pdl_flux_c_ini->dims;
$coeffs=zeroes($nf,3);

$crpix=$pdl_flux_c_ini->hdr->{CRPIX1};
$cdelt=$pdl_flux_c_ini->hdr->{CDELT1};
$crval=$pdl_flux_c_ini->hdr->{CRVAL1};
for ($i=0;$i<$nf;$i++) {
    $Av[$i]=$Av_IN;
    $d_Av[$i]=$d_Av_IN;
}





#

if ($mask_list eq "none") {
    $nmask=0;
} else {
    open(FH,"<$mask_list");
    while($line=<FH>) {
	chop($line);
	@data=split(" ",$line);
	$start_mask[$nmask]=$data[0];
	$end_mask[$nmask]=$data[1];
	$nmask++;
    }
    close(FH);
    $n_mask_org=$nmask;
    open(FH,"<$elines_mask");
    while($line=<FH>) {
	chop($line);
	if ($line !~ "#") {
	    @data=split(" ",$line);
	    $w_eline[$nline]=$data[0];
	    $nline++;
	}
    }
    close(FH);
}


#
#  We read the input spectrum
#

$n_unc=0;
$y_min=1e12;
$y_max=-1e12;
open(FH,"<$unc_file");
#open(FH,"<spec_auto.input");
$i_scale=0;
$FLUX=0;
$have_error=0;
while($line=<FH>) {
    @data=split(" ",$line);
    if ($line !~ "#") {
    $index_unc[$n_unc]=$data[0];
    $wave_unc[$n_unc]=$data[1];
    $flux_unc[$n_unc]=$data[2];
    $flux_unc_org[$n_unc]=$data[2];
    if ($#data>2) {
	$e_flux_unc[$n_unc]=sqrt(abs($data[3]));
#	if ($e_flux_unc[$n_unc]>sqrt(abs($flux_unc[$n_unc]))) {
#	    $e_flux_unc[$n_unc]=sqrt(abs($flux_unc[$n_unc]));
#	}
	$color_unc[$n_unc]=1;#$data[4];
	$have_error=1;
    } else {
	$e_flux_unc[$n_unc]=0.0;
	$color_unc[$n_unc]=1;
    }
    if ($flux_unc[$n_unc] eq "nan") {
	$flux_unc[$n_unc]=$flux_unc[$n_unc-1];
    }
    if ($flux_unc[$n_unc]<$y_min) {
	$y_min=$flux_unc[$n_unc];
    }
    if ($flux_unc[$n_unc]>$y_max) {
	$y_max=$flux_unc[$n_unc];
    }
    if ($n_unc>0) {
	if (($wave_unc[$n_unc-1]<=$wave_scale)&&($wave_unc[$n_unc]>$wave_scale)) {
	    $i_scale=$n_unc;	    
	}
    }
    $masked[$n_unc]=1;

    if ($flux_unc[$n_unc]==0) {
	$masked[$n_unc]=0;
    }

    $w_test=$wave_unc[$n_unc-1];
    for ($j=0;$j<$nmask;$j++) {
	if (($w_test>$start_mask[$j])&&($w_test<$end_mask[$j])) {
	    $masked[$n_unc]=0;	    
	}
    }
    
    if ($def==2) {
	if ($w_test<$min_wave) {
	    $masked[$n_unc]=0;	    
	}
	if ($w_test>$max_wave) {
	    $masked[$n_unc]=0;	    
	}
    }
    $flux_masked[$n_unc]=$flux_unc[$n_unc]*$masked[$n_unc];
    if (($wave_unc[$n_unc]>$min_wave)&&($wave_unc[$n_unc]<$max_wave)) {
	$FLUX=$FLUX+$flux_masked[$n_unc];
    }
    $n_unc++;
    }
}
close(FH);
#exit;

if ($def==2) {
    $y_min=$min;
    $y_max=$max;
} else {
    ($min_wave,$max_wave)=minmax(@wave_unc);
#    $min_wave=$wave_unc[0];
#    $max_wave=$wave_unc[$n_unc-1];
}

#print "$def $min_wave $max_wave\n";
#exit;

if ($def==1) {
    $y_min=$min;
    $y_max=$max;
}
#
# We check the median value
#
$median_flux=median(@flux_masked);
#open(MOUT,">median.flux");
#print MOUT "$median_flux\n";
#close(MOUT);
$dpix_unc=$wave_unc[1]-$wave_unc[0];

$max=3*$median_flux;


my $pdl_output=zeroes($n_unc,6);

#
# We create a kernel
#


#print "We start the fit...\n";

#open(OUT,">$outfile");

$med_flux=median(@flux_unc);
#$min_chi_sq=FIT_simple($redshift,$sigma,\@Av);

$chi_sq_min_now=1e12;
my $pdl_mod_SSP;
my $pdl_res_SSP;
($min_chi_sq,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($redshift,$sigma,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);


print "CUT = $med_flux $CUT_MEDIAN_FLUX\n";
print "$redshift,$sigma\n"; #<stdin>;
if ($med_flux<$ABS_MIN) {
# WHAT TO DO???
    exit;
}

if ($med_flux>$CUT_MEDIAN_FLUX) {

    
    
    
    if ($MIN_W==0) {
	$MIN_W=$min_wave;
    }
    
    if ($MAX_W==0) {
	$MAX_W=$max_wave;
    }
    
#print "$MIN_W,$MAX_W\n"; exit;

################
# REDSHIFT DETERMINATION


    $my_plot=$plot;
#    $plot=0;
    $K=0;
    open(CHI,">chi_r.txt");
    $nr=0;
    my @chi_r;
    my @red_r;
    if ($d_redshift>0) {
	$min_chi_sq=1e30;	
	
	$RED=$min_redshift;
	do {
	    $RED=$RED+$d_redshift;
#	    $chi_now=FIT_Z_SIGMA($RED,$sigma,\@Av);
	    
	    ($chi_now,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($RED,$sigma,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);

	    $chi_r[$nr]=$chi_now;
	    $red_r[$nr]=$RED;
	    
	    
#	    if (($nr>1)&&($chi_r[$nr-1]<$chi_r[$nr-2])&&($chi_r[$nr-1]<$chi_r[$nr])&&($chi_r[$nr-1]<$min_chi_sq)) {
	    if (($nr>1)&&($chi_r[$nr-1]<$min_chi_sq)) {
		$redshift=$red_r[$nr-1];	
		$min_chi_sq=$chi_r[$nr-1];	
		$K=$nr-1;
#		$RED=$max_redshift;
	    } 
	    print CHI "$nr $chi_r[$nr] $red_r[$nr] $redshift $min_chi_sq\n";
#	    print "$nr $chi_r[$nr] $red_r[$nr] $redshift $min_chi_sq\n";
	    
	    $nr++;
	}  while ($RED<$max_redshift);
#
# TWO
#
	$nr=0;
	$RED=$redshift-1.5*$d_redshift;
	$max_redshift=$redshift+1.5*$d_redshift;
	$d_redshift=0.1*$d_redshift;
#	$d_redshift=0.001;
	do {
	    $RED=$RED+$d_redshift;
	($chi_now,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($RED,$sigma,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);

#	    $chi_now=FIT($RED,$sigma,\@Av);
	    $chi_r[$nr]=$chi_now;
	    $red_r[$nr]=$RED;
	    if (($nr>1)&&($chi_r[$nr-1]<$chi_r[$nr-2])&&($chi_r[$nr-1]<$chi_r[$nr])&&($chi_r[$nr-1]<$min_chi_sq)) {
		my $a=$red_r[$nr-2];
		my $b=$red_r[$nr-1];
		my $c=$red_r[$nr];
		my $fa=$chi_r[$nr-2];
		my $fb=$chi_r[$nr-1];
		my $fc=$chi_r[$nr];
		my $den=($fc-2*$fb+$fa);
		$redshift=$c-($b-$a)*(($fc-$fb)/$den+0.5);
#$red_r[$nr-1];	
		$min_chi_sq=$chi_r[$nr-1];	
		$K=$nr-1;
	#	$RED=$max_redshift;
	    }
	    print CHI "$nr $chi_r[$nr] $red_r[$nr] $redshift $min_chi_sq\n";
	    
	    $nr++;
	}  while ($RED<$max_redshift);

	$fit_redshift=0;
	$d_redshift=0;
    } else {
	$fit_redshift=0;
    }
    close(CHI);
    

    print "REDSHIFT = $redshift\n";	#<stdin>;
    $REDSHIFT=$redshift;


# sigma DETERMINATION
    
    $K=0;
    open(CHI,">chi_s.txt");
    $nr=0;
    my @chi_s;
    my @sigma_s;
    print "D_SIGMA = $d_sigma\n";
    if ($d_sigma>0) {
	$min_chi_sq=1e30;
#    $sigma=$min_sigma;
	$SIGMA=$min_sigma;
	
	do {
	    $SIGMA=$SIGMA+$d_sigma;
#	    $chi_now=FIT_Z_SIGMA($redshift,$SIGMA,\@Av);
	($chi_now,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($redshift,$SIGMA,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);
#	    $chi_now=FIT($redshift,$SIGMA,\@Av);
	    $chi_s[$nr]=$chi_now;
	    $sigma_s[$nr]=$SIGMA;
#	if ($chi_now<$min_chi_sq) {
#	    if (($nr>1)&&($chi_s[$nr-1]<$chi_s[$nr-2])&&($chi_s[$nr-1]<$chi_s[$nr])&&($chi_s[$nr-1]<$min_chi_sq)) {
	    if ($chi_s[$nr-1]<$min_chi_sq) {
		$sigma=$sigma_s[$nr-1];	
		$min_chi_sq=$chi_s[$nr-1];    
		$K=$nr;
#		$SIGMA=$sigma;
	    }
#	    $sigma=$sigma_s[$K-1];
	    print CHI "$nr $chi_s[$nr] $sigma_s[$nr] $sigma $min_chi_sq\n";
	    $nr++;
	} while ($SIGMA<$max_sigma);

#	print "PASSED $SIGMA\n";	
	$SIGMA=$sigma-$d_sigma;
	$max_sigma=$sigma+$d_sigma;
	$d_sigma=0.33*$d_sigma;
	$nr=0;
	do {
	    $SIGMA=$SIGMA+$d_sigma;
#	    $chi_now=FIT_Z_SIGMA($redshift,$SIGMA,\@Av);
	($chi_now,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($redshift,$SIGMA,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);
#	    $chi_now=FIT($redshift,$SIGMA,\@Av);
	    $chi_s[$nr]=$chi_now;
	    $sigma_s[$nr]=$SIGMA;
#	if ($chi_now<$min_chi_sq) {
	    if (($nr>1)&&($chi_s[$nr-1]<$chi_s[$nr-2])&&($chi_s[$nr-1]<$chi_s[$nr])) {
		$sigma=$sigma_s[$nr-1];	
		$min_chi_sq=$chi_s[$nr-1];    
		$K=$nr;
		$SIGMA=$max_sigma;
	    }
	    print CHI "$nr $chi_s[$nr] $sigma_s[$nr] $sigma $min_chi_sq\n";
	    $nr++;
	} while ($SIGMA<$max_sigma);
	
	$fit_sigma=0;
	$d_sigma=0;
    } else {
	$fit_sigma=0;
    }
    close(CHI);
    print "SIGMA = $sigma\n";
} else {
#
# Below the cut!
#
    for ($i=0;$i<$nf;$i++) {
	$Av[$i]=0;
	$d_Av[$i]=0;
    }

#    $chi_now=FIT_Z_SIGMA($RED,$sigma,\@Av);
}
print "Av det\n";
# Av DETERMINATION
#($Av_IN,$d_Av_IN,$min_Av,$max_Av)=split(" ",$line);
    $K=0;
    open(CHI,">chi_Av.txt");
    $nr=0;
    my @chi_Av;
    my @Av_s;
    print "D_Av = $d_Av_IN\n";
#$d_Av_IN=$d_Av_IN*0.1;

    if ($d_Av_IN>0) {
	$min_chi_sq=1e30;
#    $sigma=$min_sigma;
	$Av_NOW=$min_Av;
	do {
	    $Av_NOW=$Av_NOW+$d_Av_IN;
	    for ($i=0;$i<$nf;$i++) {
		$Av[$i]=$Av_NOW;
	    }

#	    print "$Av_NOW start\n";
	($chi_now,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_lin($redshift,$sigma,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$min_chi_sq,$plot);
#	    print "$Av_NOW end\n";
#	    $chi_now=FIT_simple($redshift,$sigma,\@Av);
	    $chi_Av[$nr]=$chi_now;
	    $Av_s[$nr]=$Av_NOW;
	    if ($chi_Av[$nr-1]<$min_chi_sq) {
		$Av_F=$Av_s[$nr-1];	
		$min_chi_sq=$chi_Av[$nr-1];    
		$K=$nr;
	    }
	    print CHI "$nr $chi_Av[$nr] $Av_s[$nr] $Av_F $min_chi_sq\n";
	    $nr++;
	} while ($Av_NOW<$max_Av);

	$fit_Av=0;
	$d_Av_NOW=0;
    } else {
	$fit_Av=0;
    }

close(CHI);

if ($d_Av_IN==0) {
    $Av_F=$Av_IN;
}

print "AV = $Av_F\n";

for ($i=0;$i<$nf;$i++) {
    $Av[$i]=$Av_F;
}

#$call="cat chi_s.txt";
#system($call);

#exit;

#print "PASSED\n";

#$call="table_plot.pl chi_r.txt 0 1 'N.Iter' 'Chi' 101/xs";
#system($call);
#$call="table_plot.pl chi_r.txt 2 1 'Redshift' 'Chi' 102/xs";
#system($call);
#exit;

#
# 
#
if ($have_error==0) {
    @e_flux_unc=median_filter(2.345,\@res_spec);
    for ($i=0;$i<$n_unc;$i++) {
	$e_flux_unc[$i]=2.345*abs($e_flux_unc[$i]);
    }
}


#$redshift=$redshift-$d_redshift;
#print "REDSHIFT = $redshift\n";



my @new_flux_unc;
for ($i=0;$i<$n_unc;$i++) {
    $res_flux_unc[$i]=$flux_unc[$i]-$model_spec_min[$i];
    $new_flux_unc[$i]=$flux_unc[$i];
}
@median=median_filter(51,\@res_flux_unc);

open(CLEAN,">$clean_file");
for ($i=0;$i<$n_unc;$i++) {
    $flux_unc_clean[$i]=$flux_unc[$i]-$median[$i];
    print CLEAN "$i $wave_unc[$i] $flux_unc_clean[$i] $new_flux_unc[$i] $median[$i]\n";
}
close(CLEAN);



$fit_redshift=0;

#print "REDSHIFT = $redshift\n";
$redshift_abs=$redshift;
#exit;

#exit;

#print "($redshift,$sigma,\@Av) $min_chi_sq\n";
#print "----------------------\n";
$delta_chi=10;
#while (($min_chi_sq>1)&&($delta_chi>0.1)) {
$NITER=0;
$niter_tmp_max=10;


$chi_sq_min_now=1e12;
$min_chi_sq_limit=$min_chi_sq;

$n_mc=30;
$plot=$my_plot;
my $pdl_mod_JOINT;
my $pdl_res_JOINT;
my $pdl_no_gas;
my $age_min;
my $met_min;
my $Av_min;
my $age_min_mass;
my $met_min_mass;
my $Av_min_mass;
my $coeffs_cat=zeroes($nf+1,$n_mc);
while (($MIN_CHISQ>$MIN_DELTA_CHISQ)&&($NITER<$MAX_NITER)) {
    if ($NITER==1) {
	$MIN_CHISQ=1e12;
    }


    if ($med_flux>$CUT_MEDIAN_FLUX) {

	($min_chi_sq,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_mod_SSP,$pdl_res_SSP)=fit_ssp_rnd($redshift,$sigma,\@Av,$crval,$cdelt,$crpix,$nf,$n_c,$pdl_flux_c_ini,\@wave_unc,\@masked,\@e_flux_unc,\@flux_masked,$n_mc,$chi_sq_min_now,$MIN_CHISQ,$plot);
	$pdl_mod_JOINT=$pdl_mod_SSP;
	$pdl_res_JOINT=$pdl_res_SSP;
#	print "COEFFS_CAT #########\n $coeffs_cat\n ##########\n";
	open(C,">$out_coeffs_file");
	print C "# ID   AGE     MET    COEFF   Min.Coeff  log(M/L)   AV   N.Coeff   Err.Coeff\n";
	print "------------------------------------------------------------------------------\n";
	print "ID   AGE     MET    COEFF   Min.Coeff  log(M/L)   AV   N.Coeff   Err.Coeff\n";
	print "------------------------------------------------------------------------------\n";
	@age_mod=list($pdl_age_mod);
	@met_mod=list($pdl_met_mod);
	@Av_mod=list($pdl_Av);
	@ml=list($pdl_ml);
#	print "COEFFS = $coeffs\n";
	@a_coeffs=list($coeffs->slice(":,(0)"));
	@a_e_coeffs=list($coeffs->slice(":,(1)"));
	@a_min_coeffs=list($coeffs->slice(":,(2)"));
#	print "@a_e_coeffs\n";
	@a_coeffs_N=list($coeffs_N);
	my $l_age_min=0;
	my $l_met_min=0;
	my $l_Av_min=0;
	my $l_age_min_mass=0;
	my $l_met_min_mass=0;
	my $l_Av_min_mass=0;
	my $e_l_age_min=0;
	my $e_l_met_min=0;
	my $e_l_Av_min=0;
	my $e_l_age_min_mass=0;
	my $e_l_met_min_mass=0;
	my $e_l_Av_min_mass=0;
	my $ML=0;
	for ($k=0;$k<$nf;$k++) {
	    if ($a_coeffs[$k]>0) {
		$a_e_coeffs_N[$k]=$a_e_coeffs[$k]*($a_coeffs_N[$k]/$a_coeffs[$k]);
	    } else {
		$a_e_coeffs_N[$k]=0;
	    }

	    $l_age_min=$l_age_min+$a_coeffs[$k]*log10($age_mod[$k]);
	    $l_met_min=$l_met_min+$a_coeffs[$k]*log10($met_mod[$k]);
	    $l_Av_min=$l_Av_min+$a_coeffs[$k]*log10($Av_mod[$k]);	    	    

	    $l_age_min_mass=$l_age_min_mass+$ml[$k]*$a_coeffs_N[$k]*log10($age_mod[$k]);
	    $l_met_min_mass=$l_met_min_mass+$ml[$k]*$a_coeffs_N[$k]*log10($met_mod[$k]);
	    $l_Av_min_mass=$l_Av_min_mass+$ml[$k]*$a_coeffs_N[$k]*log10($Av_mod[$k]);

	    $e_l_age_min=$e_l_age_min+$a_e_coeffs[$k]*log10($age_mod[$k]);
	    $e_l_met_min=$e_l_met_min+$a_e_coeffs[$k]*log10($met_mod[$k]);
	    $e_l_Av_min=$e_l_Av_min+$a_e_coeffs[$k]*log10($Av_mod[$k]);	    	    
	    $e_l_age_min_mass=$e_l_age_min_mass+$ml[$k]*$a_e_coeffs_N[$k]*log10($age_mod[$k]);
	    $e_l_met_min_mass=$e_l_met_min_mass+$ml[$k]*$a_e_coeffs_N[$k]*log10($met_mod[$k]);
	    $e_l_Av_min_mass=$e_l_Av_min_mass+$ml[$k]*$a_e_coeffs_N[$k]*log10($Av_mod[$k]);

	    $ML=$ML+$ml[$k]*$a_coeffs_N[$k];

	    printf(C "%-2d  %-7.4f %-7.4f %-7.4f %-7.4f    %-4.4f    %-4.2f   %-7.4f %-7.4f\n",$k,$age_mod[$k],$met_mod[$k],$a_coeffs_N[$k],$a_min_coeffs[$k],log10($ml[$k]),$Av_mod[$k],$a_coeffs[$k],$a_e_coeffs[$k]);
	    if ($a_coeffs[$k]>1e-5) {
		printf("%-2d  %-7.4f %-7.4f %-7.4f %-7.4f    %-4.4f    %-4.2f   %-7.4f %-7.4f\n",$k,$age_mod[$k],$met_mod[$k],$a_coeffs_N[$k],$a_min_coeffs[$k],log10($ml[$k]),$Av_mod[$k],$a_coeffs[$k],$a_e_coeffs[$k]);
	    }
	}
	print "------------------------------------------------------------------------------\n";
	close(C);
	$age_min=10**($l_age_min);
	$met_min=10**($l_met_min);
	$Av_min=10**($l_Av_min);
	$age_min_mass=10**($l_age_min_mass/$ML);
	$met_min_mass=10**($l_met_min_mass/$ML);
	$Av_min_mass=10**($l_Av_min_mass/$ML);

	$e_age_min=$e_l_age_min*$age_min;
	$e_met_min=$e_l_met_min*$met_min;
	$e_Av_min=$e_l_Av_min*$Av_min;
	$e_age_min_mass=$e_l_age_min*$age_min_mass;
	$e_met_min_mass=$e_l_met_min*$met_min_mass;
	$e_Av_min_mass=$e_l_Av_min*$Av_min_mass;

	if ($min_chi_sq>0) {
	    $delta_chi=abs(($chi_sq_min_now-$min_chi_sq)/$min_chi_sq);
	}
    }

    $wpeak=6562;
    $Fpeak=-1e12;

######################################################################
# Fitting the emission lines
######################################################################

	my @a_fixed;
	my @a_type_fixed;
	my $n_mod_fixed=0;
    if ($ns>0) {

    $ks=0;
    $SYS_VEL=$vel_light*$REDSHIFT;
#    $call="rm ".$out_elines." > junk.junk";
    my @REN;
    my @e_REN;
    system($call);
    for ($is=0;$is<$ns;$is++) {
	if ($red_elines>0) {
	    $SYS_VEL=$vel_light*$red_elines;
	    if ($is==0) {
		$SYS_VEL_MAX=$vel_light*$red_elines+$D_SYS_VEL;
		$SYS_VEL_MIN=$vel_light*$red_elines-$D_SYS_VEL;
	    } else {
		$SYS_VEL_MAX=$vel_light*$red_elines+$D_SYS_VEL;
		$SYS_VEL_MIN=$vel_light*$red_elines-$D_SYS_VEL;
	    }
	} else {
	    $SYS_VEL=$vel_light*$REDSHIFT;
	    if ($is==0) {
		$SYS_VEL_MAX=$vel_light*$REDSHIFT+$D_SYS_VEL;
		$SYS_VEL_MIN=$vel_light*$REDSHIFT-$D_SYS_VEL;
	    } else {
		$SYS_VEL_MAX=$vel_light*$REDSHIFT+$D_SYS_VEL;
		$SYS_VEL_MIN=$vel_light*$REDSHIFT-$D_SYS_VEL;
	    }
	}
	$start_w_e=$start_w_E[$is];
	$end_w_e=$end_w_E[$is];
	$mask_e=$mask_E[$is];
	$config_e=$config_E[$is];
	$npoly_e=$npoly_E[$is];
	$mask_poly_e=$mask_poly_E[$is];
	$nmin_e=$nmin_E[$is];
	$nmax_e=$nmax_E[$is];


	
	print "CONF=$config_e\n";

	my @wave_elines;
	my @flux_elines;
	my @flux_e_elines;
	my @masked_elines;
	$n_e=0;
	for ($i=0;$i<$n_unc;$i++) {
	    if (($wave_unc[$i]>$start_w_e)&&($wave_unc[$i]<$end_w_e)) {
		$wave_elines[$n_e]=$wave_unc[$i];
		$flux_elines[$n_e]=$flux_unc_org[$i]-$pdl_mod_SSP->at($i);#$pdl_res_SSP->at($i);
		$flux_e_elines[$n_e]=sqrt($flux_unc[$i]);
		$masked_elines[$n_e]=1;
		$n_e++;
	    }
	}
	my $pdl_wave_elines=pdl(@wave_elines);
	my $pdl_flux_elines=pdl(@flux_elines);
	my $pdl_flux_e_elines=pdl(@flux_e_elines);
	my $pdl_masked_elines=pdl(@masked_elines);
	my @stats=stats($pdl_flux_elines);
	my $y_max=$stats[4]-$stats[0];
	my $def=1;


	
#	open(CNF,"<$config_e");
#	$cnf=<CNF>; chop($cnf);
	$cnf=$config_line_E[$is][0];
	($junk,$n_mod,$chi_goal,$d_chi_goal)=split(" ",$cnf);
	my $n_line=$n_line_E[$is];
	my $i_mod=1;
	for ($ii=0;$ii<$n_mod;$ii++) {
	    $cnf=$config_line_E[$is][$i_mod]; 	    $i_mod++;

#	    $cnf=<CNF>; chop($cnf);
	    $type[$ii]=$cnf;
	    for ($j=0;$j<9;$j++) {
		$cnf=$config_line_E[$is][$i_mod]; 	    $i_mod++;

#		$cnf=<CNF>; chop($cnf);
		@data=split(" ",$cnf);
		if ($type[$ii] ne "poly1d") {
		    $a[$ii][$j]=$data[0];
		    $ia[$ii][$j]=$data[1];
		} else {
		    $a[$ii][$j]=0;
		    $ia[$ii][$j]=0;
		}
		$ea[$ii][$jj]=0;
		$a0[$ii][$j]=$data[2];
		$a1[$ii][$j]=$data[3];
		$link[$ii][$j]=$data[4];
	    }
	    
	    if ($def==1) {
		$a1_max=2*$y_max*($a[$ii][2]*((2*3.1416)**0.5));
		$a0_min=0.01*1.2*$y_max*($a[$ii][2]*((2*3.1416)**0.5));
		if ($a1[$ii][1]>$a1_max) {
		    $a1[$ii][1]=$a1_max;
		    $a0[$ii][1]=$a0_min;
		}
	    }
	}
	close(CNF);
	
	#
	# Force vicitiny in the velocity
	#
	$a[0][3]=$SYS_VEL;
	$ia[0][3]=1;
	$a0[0][3]=$SYS_VEL_MIN;	
	$a1[0][3]=$SYS_VEL_MAX;

	



	my $i_ter=0;
	my $n_loops=5;
	$n_mc_e=30;
	$break=0;
	$scale_ini=0.15;
	$def=0;
	my $pdl_model=zeroes($n_e);
	my $pdl_model_cont=zeroes($n_e);
	my $pdl_model_tmp=zeroes($n_e);
	my $pdl_model_cont_tmp=zeroes($n_e);
#	print "START FIT ($n_e)\n";
	@a_out_now=copy_a($n_mod,\@a);
	@a_out_tmp=copy_a($n_mod,\@a);
        $chi_sq_now=1e12;
	my @a_results;
	$nnk=0;


	while (($i_ter<$n_loops)&&($break==0)) {
	    my $chi_iter=$chi_sq_now;
#	    srand(localtime);
	    $chi_single=0;
	    ($chi_sq_now,$pdl_a,$pdl_model_tmp,$pdl_model_cont_tmp)=fit_elines_grad_rnd($pdl_wave_elines,$pdl_flux_elines,$pdl_flux_e_elines,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out_tmp,\@ia,\@a0,\@a1,\@link,$n_mc_e,$pdl_masked_elines,$def,$scale_ini,4);
	    @a_out_now=copy_a_pdl($n_mod,$pdl_a);
	    if ($chi_sq_now<$chi_iter) {
		@a_out_tmp=copy_a($n_mod,\@a_out_now);
		@a_results=copy_a_results($n_mod,$nnk,\@a_out_now,\@a_results);    
		$pdl_model=$pdl_model_tmp;
		$pdl_model_cont=$pdl_model_cont_tmp;
		$nnk++;
		$i_ter++;
	    } else {
		my $rnd_a=random(10);
		@a_out_now=copy_a($n_mod,\@a_out_now);
	    }
	}
	my $out_ps_now="fit_".$outfile.".".$start_w_e."_".$end_w_e;
	my $title="[$start_w_e,$end_w_e]";
	plot_results($plot,$pdl_wave_elines,pdl($pdl_flux_elines,$pdl_model,($pdl_flux_elines-$pdl_model)),$out_ps_now,$title);

	print "----------------------------------------\n";
	@a_final=mean_a_results_last($n_mod,$nnk,\@a_results);
	print_a_final($n_mod,\@a_final,\@type,$chi_sq_now);
	$out_fit_spectra=$out_elines;
	print_a_final_file_add($n_mod,\@a_final,\@type,$chi_sq_now,$out_fit_spectra);

#	my $iiF,$jF;
#	for ($iiF=0;$iiF<$n_mod;$iiF++) {   
#	    $a_type_fixed[$iiF+$n_mod_fixed]=$a_type[$iiF];
#	for ($jF=0;$jF<9;$jF++) {
#	    $a_final_fixed[$iiF+$n_mod_fixed][$jF]=$a_final[$iiF][$jF];
#	}
#	}
#	$n_mod_fixed=$n_mod_fixed+$n_mod;
	
	($n_mod_fixed,$junk_a_fixed,$junk_a_type_fixed)=add_a_results_elines($n_mod,\@a_final,\@type,$n_mod_fixed,\@a_fixed,\@a_type_fixed);
	@a_fixed=@$junk_a_fixed;
	@a_type_fixed=@$junk_a_type_fixed;



	$nmin_e=int(0.1*$n_unc);
	$nmax_e=int(0.9*$n_unc);

###############################
# Low order polynomical!
#	$call="spec2img.pl sub_spec.txt res_spec.fits 1";
#	system($call);
#	$call="polyfit_mask.pl res_spec.fits 6 res_poly.fits ".$mask_poly_e." ".$nmin_e." ".$nmax_e." /null";
#	system($call);
#	$call="imarith.pl res_spec.fits - res_poly.fits new_res.fits";
#	system($call);
#	$call="img2spec.pl new_res.fits 0 new_res.txt";
#	system($call);

	$out_fit_now=$out_fit.".".$start_w_e."_".$end_w_e.".ps/CPS";
	$box=int($sigma*6);
	
	print "DONE FIT ELINES CONFIG $is\n";
    }


#
# We create a FIXED model of the emission lines
#

    my $pdl_model_elines=zeroes($n_unc);
    my $pdl_model_cont=zeroes($n_unc);
    my $pdl_wave_elines=pdl(@wave_unc);
    for ($i=0;$i<$n_mod_fixed;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave_elines,$i,\@a_type_fixed,\@a_fixed);
	$pdl_model_elines=$pdl_model_elines+$pdl_tmp;
    }



    #
    # We remove the gas before a new iteration
    #
    
#    @model_join
    for ($i=0;$i<$n_unc;$i++) {
	$flux_unc[$i]=$flux_unc_org[$i]-$pdl_model_elines->at($i);
    }

    $pdl_mod_JOINT=$pdl_mod_SSP+$pdl_model_elines;
    $pdl_res_JOINT=$pdl_res_SSP-$pdl_model_elines;
    $pdl_no_gas=pdl(@flux_unc);

#    $pdl_output=pdl(
    # Copy output!
    $t=$pdl_output->slice(":,0");
    $t .=pdl(@flux_unc_org);
    $t=$pdl_output->slice(":,1");
    $t .=$pdl_mod_SSP;
    $t=$pdl_output->slice(":,2");
    $t .=$pdl_mod_JOINT;
    $t=$pdl_output->slice(":,3");
    $t .=$pdl_res_SSP;
    $t=$pdl_output->slice(":,4");
    $t .=$pdl_res_JOINT;
    $t=$pdl_output->slice(":,5");
    $t .=pdl(@flux_unc_org)-$pdl_model_elines;

#    plot_results($plot,$pdl_wave_elines,pdl(pdl(@flux_unc_org),$pdl_mod_JOINT,$pdl_res_JOINT,$pdl_mod_SSP,$pdl_res_SSP),$out_ps);
    my $title="X=$chi_sq_now T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) Av=@Av z=$redshift sigma=$sigma";
    plot_results($plot,$pdl_wave_elines,$pdl_output,$out_ps,$title);

    }
    print "I.Iter = $NITER DONE\n";
    $NITER++;

}

# Write output file
$h = {NAXIS=>2, NAXIS1=>$n_unc, NAXIS=>6, COMMENT=>"OUTPUT auto_ssp_elines_rnd.pl FITs"};
$$h{CRVAL1}=$wave_unc[0];
$$h{CDELT1}=$wave_unc[1]-$wave_unc[0];
$$h{CRPIX1}=1;
$pdl_output->sethdr($h);
$pdl_output->wfits($out_fit);
################################

print "--------------------------------------------------------------\n";
my $pdl_masked=pdl(@masked);
my $pdl_chi_now=(($pdl_masked*$pdl_res_JOINT)**2)/((pdl(@e_flux_unc))**2);
$pdl_chi_now->inplace->setnantobad;
$pdl_chi_now->inplace->setbadtoval(0);
#print "$pdl_chi_now\n";
my $chi_joint=sum($pdl_chi_now);
$chi_joint=($chi_joint/($n_unc-$n_mod_fixed-$nf-1))**0.5;

$rms=sigma(list($pdl_masked*$pdl_res_JOINT));

my $title="X=$chi_joint T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) Av=@Av z=$redshift sigma=$sigma";
plot_results($plot,pdl(@wave_unc),$pdl_output,$out_ps,$title);

    print "MSP CHISQ=$chi_joint AGE=$age_min+-$e_age_min MET=$met_min+-$e_met_min AV=$Av_min+-$e_Av_min REDSHIFT=$redshift+-$e_reshift SIGMA_DISP=$sigma+-$e_sigma RMS=$rms MED_FLUX=$med_flux AGE_mass=$age_min_mass+-$e_age_min_mass MET_mass=$met_min_mass+-$e_met_min_mass\n";

#close(OUT);



open(FH,">$outfile");
print FH "#MIN_CHISQ age_min met_min Av redshift sigma median_FLUX redshift_ssp med_flux rms_residual age_min_mass met_min_mass SYS_VEL $back_list \n";
if ($chi_joint==0) {
    $chi_joint=1;
}
print FH "$chi_joint $age_min $met_min $Av_min $redshift $sigma $FLUX $redshift_abs $med_flux $rms $age_min_mass $met_min_mass $SYS_VEL\n";
close(FH);

$sec_end=print_time();

$sec_total=$sec_end-$sec_ini;
print "# SECONDS = $sec_total\n";


#
# Write the output!
# 
#


exit;


#
# We save the residual spectrum
#
#print "*** n_unc=$n_unc ****** END ******";<stdin>;
open(OUT,">org_spec.txt");
open(OUT2,">org_phot.txt");
for ($j=0;$j<$n_unc;$j++) {
    if ($color_unc[$j]==1) {
	print OUT "$index_unc[$j] $wave_unc[$j] $flux_unc[$j]\n";
    } else {
	print OUT2 "$index_unc[$j] $wave_unc[$j] $flux_unc[$j] $e_flux_unc[$j]\n";
    }
}
close(OUT2);
close(OUT);
#print "EXIT"; exit;

open(OUT,">res_spec.txt");
for ($j=0;$j<$n_unc;$j++) {
    if ($color_unc[$j]==1) {
	print OUT "$index_unc[$j] $wave_unc[$j] $res_spec[$j]\n";
    }
}
close(OUT);



$NBOX=int(9*$sigma);
$N1=int(0.05*$n_unc);
$N2=int(0.95*$n_unc);
$call="smooth_spec_clip.pl res_spec.txt tmp_spec.txt ".$NBOX." /null 2 ".$N1." ".$N2;
system($call);
#print "$call\n";





$call="spec_arith.pl res_spec.txt - tmp_spec.txt gas_spec.txt";
system($call);


open(OUT,">model_spec.txt");
for ($j=0;$j<$n_unc;$j++) {
    if ($color_unc[$j]==1) {
	print OUT "$index_unc[$j] $wave_unc[$j] $model_spec_min[$j]\n";
    }
}
close(OUT);

#
#
$NF=0;
open(MOD,"<out_mod_res.fit_spectra");
@model_joint=@model_spec_min;
@res_joint=@res_spec;
open(OUT,">mod_joint_spec.txt");
open(OUT_RES,">res_joint_spec.txt");
while($line=<MOD>) {
    chop($line);
    @data=split(" ",$line);
    $WMOD=$data[0];
    $FMOD=$data[2];
    $model_joint[$NF]=$model_spec_min[$NF]+$FMOD;
    $res_joint[$NF]=$flux_unc[$NF]-$model_joint[$NF];
    print OUT "$index_unc[$NF] $wave_unc[$NF] $model_joint[$NF]\n";
    print OUT_RES "$index_unc[$NF] $wave_unc[$NF] $res_joint[$NF]\n";
    $NF++;
}
close(OUT);
close(OUT_RES);
close(MOD);


open(OUT,">no_gas_spec.txt");
for ($j=0;$j<$n_unc;$j++) {
    my $dat=$model_spec_min[$j]+$res_joint[$j];
    if ($color_unc[$j]==1) {
	print OUT "$index_unc[$j] $wave_unc[$j] $dat\n";
    }
}
close(OUT);

#while (($MIN_CHISQ>$MIN_DELTA_CHISQ)&&($NITER<$MAX_NITER)) {

if (($wave_norm>0)&&($MAX_NITER>0)) {
#	print "*************** FIT_SINGLE SSP ******************\n";
	$MIN_CHISQ=1e12;
#	$plot=$plot_old;
#	print "Press Enter"; <stdin>;
	
	$min_chi_sq=FIT_SINGLE($redshift,$sigma,\@Av);
	$rms_single=sigma(@res_clean);
	$med_flux_single=median(@flux_clean);
	print "SSP CHISQ=$min_chi_sq AGE=$age_min MET=$met_min AV=@Av REDSHIFT=$redshift SIGMA_DISP=$sigma RMS=$rms_single MED_FLUX=$med_flux_single\n";
#	print "Press Enter"; <stdin>;
	open(FH,">$out_single");
	print FH "#MIN_CHISQ age_min met_min Av redshift sigma median_FLUX redshift_ssp  med_flux rms_residual age_min_mass met_min_mass $new_back_file \n";
	print FH "$min_chi_sq $age_min $met_min $Av_min $redshift $sigma $FLUX $red_abs $red_em $med_flux_single  $rms_single  $age_min $met_min\n";
	close(FH);


	system($call);
}
#    <stdin>;

    
    



$time2=get_time();

print "START $time1\n";
print "END   $time2\n";


$sec_end=print_time();

$sec_total=$sec_end-$sec_ini;
print "# SECONDS = $sec_total\n";
exit;

sub print_time {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    print "# TIME $sec $min $hour $mday $mon $year $wday $yday $isdst\n";
    print LOG "# TIME $sec $min $hour $mday $mon $year $wday $yday $isdst\n";
    my $sec_now=$hour*3600+$min*60+$sec;
    return $sec_now;
}

sub FIT {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;
    my $k,$j,$i,$iter;

    my $iter_max=5;
    my $last_chi=1e12;
    
    for ($i=0;$i<$nf;$i++) {
	if ($Av[$i]<0) {
	    $Av[$i]=0;
	}
    }


    for ($j=0;$j<$n_c;$j++) {
	$wave_c[$j]=($crval+$cdelt*($j+1-$crpix))*(1+$redshift);
	if ($j>0) {
	    $dpix_c_val[$j]=$wave_c[$j]-$wave_c[$j-1];	
	}
	
    }
    $dpix_c_val[0]=$dpix_c_val[1];
$dpix_c=$wave_c[1]-$wave_c[0];

$rsigma=$sigma/$dpix_c;

    my @age_mod;
    my @met_mod;
for ($iii=0;$iii<$nf;$iii++) {
    $header="NAME".$iii;
    $name[$iii]=$pdl_flux_c_ini->hdr->{$header};
    $name_min=$name[$iii];
    $name_min =~ s/spec_ssp_//;
    $name_min =~ s/.spec//;    
    $name_min =~ s/.dat//;
    ($AGE,$MET)=split("_",$name_min);
    if ($AGE =~ "Myr") {
	$age=$AGE;
	$age =~ s/Myr//;
	$age=$age/1000;
    } else {
	$age=$AGE;
	$age =~ s/Gyr//;
    }
    $met=$MET;
    $met =~ s/z/0\./;

    $age_mod[$iii]=$age;
    $met_mod[$iii]=$met;
    $header="NORM".$iii;

 $val_ml=$pdl_flux_c_ini->hdr->{$header};
    if ($val_ml!=0) {
	$ml[$iii]=1/$val_ml;
    } else {
	$ml[$iii]=1;
    }
    
    $box=int(3*$rsigma);
    if ($box<3) {
	$box=3;
    }
    $kernel=zeroes(2*$box+1);
    $norm=0;
    $flux_c[$i]=0;
    for ($j=0;$j<2*$box+1;$j++) {
	$gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	set($kernel,$j,$gaus);
	$norm=$norm+$gaus;
    }
    $kernel=$kernel/$norm;
    

    $pdl_flux_c_conv = conv2d $pdl_flux_c_ini,$kernel;



    $pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");


     my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c), $pdl_flux_c);
    ($n_c_out)=$out_spec_pdl->dims;


    for ($i=0;$i<$n_unc;$i++) {
	$val=$out_spec_pdl->at($i);
	if ($val eq "nan") {
	    $val=0;
	}
	$model[$i][$iii]=$val*$masked[$i];
	$model_no_mask[$i][$iii]=$val;#*$masked[$i];

	if ($masked[$i]>0) {
	    $error[$i]=0.01*abs($e_flux_unc[$i]);
	} else {
	    $error[$i]=0;
	}
	for ($j=0;$j<$nline;$j++) {
	    if (abs($wave_unc[$i]-$w_eline[$j]*(1+$redshift))<(3*$rsigma)) {
#		    $error[$i]=0.1*$e_flux_unc[$i];
#		    $error[$i]=30;#*$e_flux_unc[$i];
		}
	}
    }
}

    $pdl_model=zeroes($n_unc,$nf);
    $pdl_model_no_mask=zeroes($n_unc,$nf);
    $pdl_error=zeroes($n_unc);
    my $pdl_masked=pdl(@masked);
    for ($j=0;$j<$nf;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
	    $wave_res=$wave_unc[$i]/(1+$redshift);
	    $dust_rat=A_l(3.1,$wave_res);
	    $dust=10**(-0.4*$Av[$j]*$dust_rat);  
	    $val=$model[$i][$j]*$dust;
	    set($pdl_model,$i,$j,$val);
	    $val_no_mask=$model_no_mask[$i][$j]*$dust;
	    set($pdl_model_no_mask,$i,$j,$val_no_mask);
	    $e_val=$error[$i];
	    $val_now=$e_flux_unc[$i];
	    if ($val_now==0) {
		$val_now=1;
	    }
	    set($pdl_error,$i,1/(abs($val_now)**2));
	}
    }
#
# We fit
#
    $pdl_flux_masked=pdl(@flux_masked);
    for ($j_mc=0;$j_mc<$n_mc;$j_mc++) {
	$C_left=1;
	$pdl_random=random($nf);
	# New converging criteria
#	$pdl_random=$pdl_C_last*
	for ($j=0;$j<$nf;$j++) {
	    $val=$pdl_random->at($j);
	    $val=$val*$C_left;
	    set($pdl_random,$j,$val);
	    $C_left=$C_left-$val;	    
	}
	$SUM=$pdl_random->sum;
	$pdl_random=$pdl_random/$SUM;
	$SUM=$pdl_random->sum;
	$pdl_C_now=zeroes($nf);

	$h_nf=1;
	for ($jj=0;$jj<$nf;$jj=$jj+$h_nf) {
	    $y_model_now=zeroes($n_unc);
	    $y_model_no_mask_now=zeroes($n_unc);
	    for ($j=0;$j<$nf;$j++) {
		$pdl_model_j=$pdl_model->slice(":,($j)");
		$pdl_model_no_mask_j=$pdl_model_no_mask->slice(":,($j)");
		$J=$j+$jj;
		if ($J>=$nf) {
		    $J=$J-$nf;
		}
		$val=$pdl_random->at($J);
		set($pdl_C_now,$j,$val);
		$y_model_now=$y_model_now+$val*$pdl_model_j;		
		$y_model_no_mask_now=$y_model_no_mask_now+$val*$pdl_model_no_mask_j;		
	    }
	    @dim_now=$y_model_now->dims;

	    $j1=int(0.4*$n_unc);
	    $j2=int(0.6*$n_unc);
	    my @a_norm;
	    my @b_norm;
	    my $j_a;
	    for ($j=$j1;$j<$j2;$j++) {
		$a_norm[$j_a]=$flux_unc[$j];
		$b_norm[$j_a]=$y_model_now->at($j);
		$j_a++;
	    }
	    $med_norm=median(@a_norm)/median(@b_norm);
	    $y_model_now=$y_model_now*$med_norm;
	    $y_model_no_mask_now=$y_model_no_mask_now*$med_norm;

##############################
# CHISQ VALUE
	    $chi=0;
	    $chi2=0;
	    $NFREE=0;
	    for ($j=0;$j<$n_unc;$j++) {
		$out_spec[$j]=$y_model_now->at($j,0);
		$chi_sec[$j]=0;
		if (($flux_unc[$j]!=0)&&($out_spec[$j]!=0)) {
		    $chi=$chi+$masked[$j]*(($flux_masked[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
		    if ($have_error==0) {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs($out_spec[$j]);
		    } else {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/($e_flux_unc[$j])**2;
		    }
		    $NFREE++;
		}
	    }
	    
	    $chi_sq=$chi;
	    if ($NFREE>0) {
		$chi_sq=($chi_sq/($NFREE))**0.5;
	    }

	    if ($chi_sq<$chi_sq_min_now) {
		$chi_sq_min_now=$chi_sq;
		$t=$coeffs->slice(":,0");
		$t .= $pdl_C_now;
		$y_model_end=$y_model_now;
		$y_model_no_mask_end=$y_model_no_mask_now;
	    }
	}
    }
    print ".";


    for ($j=0;$j<$n_unc;$j++) {
	$out_spec[$j]=$y_model_no_mask_end->at($j,0);
	$model_spec[$j]=$out_spec[$j];
	$res_spec[$j]=$flux_unc[$j]-$model_spec[$j];	    
	$model_spec_min[$j]=$model_spec[$j];
    }
    $chi_sq=$chi_sq_min_now;








    if ($chi_sq<$MIN_CHISQ) {
	$MIN_CHISQ=$chi_sq;
	print "\n";
	for ($j=0;$j<$n_unc;$j++) {
            $wave_res=$wave_unc[$j]/(1+$redshift);
            $dust_rat=A_l(3.1,$wave_res);
            $norm_C=0;
            $norm_C_mass=0;
            for ($k=0;$k<$nf;$k++) {
                $dust=10**(-0.4*$Av[$k]*$dust_rat);
                $C=$coeffs->at($k,0);
                $norm_C=$norm_C+$C;
                $norm_C_mass=$norm_C_mass+$C*$ml[$k];
            }
	}
	$age_min=0;
	$met_min=0;
	$Av_min=0;
	$age_min_mass=0;
	$met_min_mass=0;
	$Av_min_mass=0;
	open(C,">coeffs.out");
	print C "ID   AGE     MET    COEFF   Norm.Coeff  M/L   AV\n";
	print "---------------------------------------------\n";
	print "ID AGE     MET    COEFF   Norm.Coeff   M/L    AV\n";
	print "---------------------------------------------\n";

	$norm_C=0;
	$norm_C_mass=0;
	for ($k=0;$k<$nf;$k++) {
	    $dust=10**(-0.4*$Av[$k]*$dust_rat);
	    $C=$coeffs->at($k,0);
	    $norm_C=$norm_C+$C;
	    $norm_C_mass=$norm_C_mass+$C*$ml[$k];
	}


	for ($k=0;$k<$nf;$k++) {
	    $C=$coeffs->at($k,0);
	    if ($norm_C>0) {
		$age_min=$age_min+$C*log10($age_mod[$k])/$norm_C;
		$met_min=$met_min+$C*$met_mod[$k]/$norm_C;
		$Av_min=$Av_min+$C*$Av[$k]/$norm_C;	    
		$CN=$C/$norm_C;
	    }
	    $C_now=$C*$med_norm;
	    printf(C "%-2d %-7.4f %-7.4f %-7.4f %-7.4f %-10.2f %-4.2f\n",$k,$age_mod[$k],$met_mod[$k],$C_now,$CN,$ml[$k],$Av[$k]);
	    if ($C>1e-5) {
		printf("%-2d %-7.4f %-7.4f %-7.4f %-7.4f %-10.2f %-4.2f\n",$k,$age_mod[$k],$met_mod[$k],$C_now,$CN,$ml[$k],$Av[$k]);
	    }
	    if ($norm_C_mass>0) {
		$age_min_mass=$age_min_mass+$C*log10($ml[$k]*$age_mod[$k])/$norm_C_mass;
		$met_min_mass=$met_min_mass+$C*$ml[$k]*$met_mod[$k]/$norm_C_mass;
		$Av_min_mass=$Av_min_mass+$C*$ml[$k]*$Av[$k]/$norm_C_mass;
	    }

	}
	print "---------------------------------------------\n";
	close(C);
	$age_min=10**($age_min);
	$age_min_mass=10**($age_min_mass);
	
    }
    $name=$unc_file.", ";
    $scale="1";
    
    if ($plot==1) {

	pgbegin(0,$dev_plot,1,1);
	pgsfs(1.2);
	pgscf(2);             # Set character font
	pgslw(2);             # Set line width
	pgsch(1.2);           # Set character height
	pgenv($min_wave,$max_wave,$y_min,$y_max,0,0);
	pgsch(0.9);           # Set character height
	my $mean_Av=mean(@Av);
	pglabel("Wavelength","Counts","X=$chi_sq ($chi_joint) T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) Av=$mean_Av z=$redshift sigma=$sigma");
	pgsch(0.5);
	pgsci(1);
	pgline($n_unc,\@wave_unc,\@flux_unc);    
	pgsci(2);
	pgline($n_unc,\@wave_unc,\@out_spec);    
	pgsci(8);
	pgline($n_unc,\@wave_unc,\@model_spec_min);    
	pgsci(5);
	pgline($n_unc,\@wave_unc,\@res_spec);    

	if ($nc>0) {
	    pgsci(6);
	    pgline($nc,\@wave_clean,\@res_clean);    
	}
	pgsci(1);
	pgsci(13);
	pgsls(2);
	$j_plot=0;



	for ($j=0;$j<$nf;$j++) {
	    $C=$coeffs->at($j,0);
	    if ($C>0) {
		pgsci(13);
		my $model_now=$pdl_model->slice(",($j)");

		$model_now=$C*$model_now;
		@a_model_now=list($model_now);
		pgline($n_unc,\@wave_unc,\@a_model_now);    

	    }
	    if ($norm_C != 0) {
		$CN=$C/$norm_C;
	    } else {
		$CN=$C;
	    }
	    if ($C>0) {
		$pos_x=$min_wave*1.02;
		$pos_y=$y_max-0.05*($j_plot+1)*($y_max-$y_min);
		pgsch(0.8);           # Set character height
		if ($C>0) {
		    pgsci(1);
		} else {
		pgsci(15);
		}
		$CN=apr($CN);
		pgptxt($pos_x,$pos_y,0,0,"$CN $age_mod[$j] $met_mod[$j]");
		$j_plot++;
#	    print "$pos_x $pos_y\n"; <stdin>;
	    }
		pgsch(1.2);           # Set character height

	}
	pgsls(1);


	pgsci(1);


	pgclose;
	pgend;

    }

    return $chi_sq;
}


sub fit_ssp_rnd {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;
    my $crval=$_[3];
    my $cdelt=$_[4];
    my $crpix=$_[5];
    my $nf=$_[6];
    my $n_c=$_[7];
    my $pdl_flux_c_ini=$_[8];
    my $junk=$_[9];
    my @wave_unc=@$junk;
    my $junk=$_[10];
    my @masked=@$junk;
    my $junk=$_[11];
    my @e_flux_unc=@$junk;
    my $junk=$_[12];
    my @flux_masked=@$junk;
    my $n_mc=$_[13];
    my $chi_sq_min_now=$_[14];
    my $MIN_CHISQ=$_[15];
    my $plot=$_[16];
    my $k,$j,$i,$iter;
    my $iii,$jj;
    my $iter_max=5;
    my $last_chi=1e12;

    my $coeffs=zeroes($nf,3);    
    my $coeffs_N=zeroes($nf,1);    
    my $coeffs_NM=zeroes($nf,1);    

    my $coeffs_cat=zeroes($nf+1,$n_mc*$n_mc);

    my $pdl_model_spec_cat=zeroes($n_unc,$n_mc*$n_mc);

    for ($i=0;$i<$nf;$i++) {
	if ($Av[$i]<0) {
	    $Av[$i]=0;
	}
    }

    my @wave_c;
    my @dpix_c_val;
    for ($j=0;$j<$n_c;$j++) {
	$wave_c[$j]=($crval+$cdelt*($j+1-$crpix))*(1+$redshift);
	if ($j>0) {
	    $dpix_c_val[$j]=$wave_c[$j]-$wave_c[$j-1];	
	}
	
    }

    $dpix_c_val[0]=$dpix_c_val[1];
    $dpix_c=$wave_c[1]-$wave_c[0];

    my $rsigma=$sigma/$dpix_c;

    my @name;
    my @age_mod,@met_mod;
    my @ml;
    my @flux_c;
    my $pdl_flux_c_conv;


    my @model;
    my @model_no_mask;
    
    my $y_model_end;
    my $y_model_no_mask_end;


    my @MED_NORM;

	my $age_min;
	my $met_min;
	my $Av_min;
	my $age_min_mass;
	my $met_min_mass;
	my $Av_min_mass;

    for ($iii=0;$iii<$nf;$iii++) {
	my $header="NAME".$iii;
	$name[$iii]=$pdl_flux_c_ini->hdr->{$header};
	my $name_min=$name[$iii];
	$name_min =~ s/spec_ssp_//;
	$name_min =~ s/.spec//;    
	$name_min =~ s/.dat//;
	my $AGE,$MET;
	my $age,$met;
	($AGE,$MET)=split("_",$name_min);
	if ($AGE =~ "Myr") {
	    $age=$AGE;
	    $age =~ s/Myr//;
	    $age=$age/1000;
	} else {
	    $age=$AGE;
	    $age =~ s/Gyr//;
	}
	$met=$MET;
	$met =~ s/z/0\./;	
	$age_mod[$iii]=$age;
	$met_mod[$iii]=$met;
	$header="NORM".$iii;	
	my $val_ml=$pdl_flux_c_ini->hdr->{$header};
	if ($val_ml!=0) {
	    $ml[$iii]=1/$val_ml;
	} else {
	    $ml[$iii]=1;
	}
	
	my $box=int(3*$rsigma);
	if ($box<3) {
	    $box=3;
	}
	my $kernel=zeroes(2*$box+1);
	my $norm=0;
	$flux_c[$i]=0;
	for ($j=0;$j<2*$box+1;$j++) {
	    $gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	    set($kernel,$j,$gaus);
	    $norm=$norm+$gaus;
	}
	$kernel=$kernel/$norm;
        $pdl_flux_c_conv = conv2d $pdl_flux_c_ini,$kernel;
	my $pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");
	my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c), $pdl_flux_c);
	my ($n_c_out)=$out_spec_pdl->dims;
	my @error;

	for ($i=0;$i<$n_unc;$i++) {
	    my $val=$out_spec_pdl->at($i);
	    if ($val eq "nan") {
		$val=0;
	    }
	    $model[$i][$iii]=$val*$masked[$i];
	    $model_no_mask[$i][$iii]=$val;#*$masked[$i];
	    if ($masked[$i]>0) {
		$error[$i]=0.01*abs($e_flux_unc[$i]);
	    } else {
		$error[$i]=0;
	    }

	}
    }

#    print "Av = @Av\n";


    my $pdl_model=zeroes($n_unc,$nf);
    my $pdl_model_no_mask=zeroes($n_unc,$nf);
    my $pdl_error=zeroes($n_unc);
    my $pdl_masked=pdl(@masked);
    for ($j=0;$j<$nf;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
	    my $wave_res=$wave_unc[$i]/(1+$redshift);
	    my $dust_rat=A_l(3.1,$wave_res);
	    my $dust=10**(-0.4*$Av[$j]*$dust_rat);  
	    my $val=$model[$i][$j]*$dust;
	    set($pdl_model,$i,$j,$val);
	    my $val_no_mask=$model_no_mask[$i][$j]*$dust;
	    set($pdl_model_no_mask,$i,$j,$val_no_mask);
	    my $e_val=$error[$i];
	    my $val_now=$e_flux_unc[$i];
	    if ($val_now==0) {
		$val_now=1;
	    }
	    set($pdl_error,$i,1/(abs($val_now)**2));
	}
    }
#
# We fit
#
    my $pdl_flux_masked=pdl(@flux_masked);
    my $j_mc;
    my $ini_cat=0;
    my $pdl_C_flast=ones($nf);
    for ($j_mc=0;$j_mc<$n_mc;$j_mc++) {
	my $C_left=1;
#	srand(time() | $$);
	my $pdl_random=random($nf);

	for ($j=0;$j<$nf;$j++) {
	    my $val=$pdl_random->at($j);
	    my $val=$val*$C_left;
	    set($pdl_random,$j,$val);
	    $C_left=$C_left-$val;	    
	}
	my $SUM=$pdl_random->sum;
	my $pdl_random=$pdl_random/$SUM;
	my $SUM=$pdl_random->sum;
	my $pdl_C_now=zeroes($nf);

	my $h_nf=1;




	for ($jj=0;$jj<$nf;$jj=$jj+$h_nf) {
	    my $y_model_now=zeroes($n_unc);
	    my $y_model_no_mask_now=zeroes($n_unc);
	    for ($j=0;$j<$nf;$j++) {
		my $pdl_model_j=$pdl_model->slice(":,($j)");
		my $pdl_model_no_mask_j=$pdl_model_no_mask->slice(":,($j)");
		my $J=$j+$jj;
		if ($J>=$nf) {
		    $J=$J-$nf;
		}
		my $val=$pdl_random->at($J);
		#
		# We center in a good solution
		#

		set($pdl_C_now,$j,$val);
		$y_model_now=$y_model_now+$val*$pdl_model_j;		
		$y_model_no_mask_now=$y_model_no_mask_now+$val*$pdl_model_no_mask_j;		
	    }
	    my @dim_now=$y_model_now->dims;

	    my $j1=int(0.4*$n_unc);
	    my $j2=int(0.6*$n_unc);
	    my @a_norm;
	    my @b_norm;
	    my $j_a;
	    for ($j=$j1;$j<$j2;$j++) {
		$a_norm[$j_a]=$flux_unc[$j];
		$b_norm[$j_a]=$y_model_now->at($j);
		$j_a++;
	    }
#	    print "A=@a_norm\n";
#	    print "B=@b_norm\n";
	    my $med_norm=median(@a_norm)/median(@b_norm);
	    $MED_NORM[$jj]=$med_norm;
	    $y_model_now=$y_model_now*$med_norm;
	    $y_model_no_mask_now=$y_model_no_mask_now*$med_norm;

##############################
# CHISQ VALUE
	    my $chi=0;
	    my $chi2=0;
	    my $NFREE=0;
	    my @out_spec;
	    my @chi_sec;
	    for ($j=0;$j<$n_unc;$j++) {
		$out_spec[$j]=$y_model_now->at($j,0);
		$chi_sec[$j]=0;
		if (($flux_unc[$j]!=0)&&($out_spec[$j]!=0)) {
		    $chi=$chi+$masked[$j]*(($flux_masked[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
		    if ($have_error==0) {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs($out_spec[$j]);
		    } else {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/($e_flux_unc[$j])**2;
		    }
		    $NFREE++;
		}
	    }
	    
	    my $chi_sq=$chi;
	    if ($NFREE>0) {
		$chi_sq=($chi_sq/($NFREE))**0.5;
	    }

#	    print "jj=$jj\n";
	    if ($chi_sq<$chi_sq_min_now) {
#		print "INI_CAT $ini_cat/$n_mc\n";

		$chi_sq_min_now=$chi_sq;
		my $t=$coeffs->slice(":,0");
		$t .= $pdl_C_now;
		$y_model_end=$y_model_now;
		$y_model_no_mask_end=$y_model_no_mask_now;
		my $nf_1=$nf-1;
		my $t=$coeffs_cat->slice("0:$nf_1,($ini_cat)");
		$t .= $pdl_C_now->slice(":,(0)");
		set($coeffs_cat,$nf,$ini_cat,$chi_sq);
		my $t=$pdl_model_spec_cat->slice(":,($ini_cat)");
		$t .= $y_model_no_mask_end;
		$ini_cat++;
	    }
#	    print "jj=$jj END\n";
	}
#    print "PASO, $j_mc/$n_mc***\n";
    }

 #   print "Termino Bucle\n";

#    print "$j_mc/$n_mc\n";


#
# We construct the average model
# and the average coefficients
#

    my @model_spec;
    my @res_spec;
    my @model_spec_min;
    my @out_spec_now;
    my $SUM_W=0;
    my @out_coeffs;
    my @out_coeffs_e;
    my $N_CASES=0;
    for ($J=0;$J<$ini_cat;$J++) {
	my $CHI=$coeffs_cat->at($nf,$J);
	if ($CHI<2*$chi_sq_min_now) {
	    for ($j=0;$j<$n_unc;$j++) {
		$out_spec_now[$j]=$out_spec_now[$j]+($pdl_model_spec_cat->at($j,$J))/$CHI;		
	    }
#	    my $pdl_coeff_now=$coeffs_cat->slice(":,$J");
	    for ($j=0;$j<$nf;$j++) {
		$out_coeffs[$j][$N_CASES]=$coeffs_cat->at($j,$J);#/$CHI;
	    }
	    $N_CASES++;
	    $SUM_W=$SUM_W+1/$CHI;
	}
    }

    for ($j=0;$j<$n_unc;$j++) {
	$model_spec[$j]=$out_spec_now[$j]/$SUM_W;
	$out_spec[$j]=$out_spec_now[$j]/$SUM_W;
	$res_spec[$j]=$flux_unc[$j]-$model_spec[$j];	    
	$model_spec_min[$j]=$model_spec[$j];
    }

    $min_coeffs=$coeffs;

    for ($j=0;$j<$nf;$j++) {
	my @tmp;
	for ($J=0;$J<$N_CASES;$J++) {
	    $tmp[$J]=$out_coeffs[$j][$J];#/$SUM_W;
#	    print "$j,$J,$tmp[$J]\n";
	}
	my $val=mean(@tmp);
	my $sigma=0.5*sigma(@tmp);
	my $old_val=$coeffs->at($j,0);
	set($coeffs,$j,0,$val);
	set($coeffs,$j,1,$sigma);
	set($coeffs,$j,2,$old_val);
    }

    my $chi_sq=$chi_sq_min_now;
	
    my $pdl_model_spec_cat=zeroes($n_unc,$n_mc);




    if ($chi_sq<$MIN_CHISQ) {
	$MIN_CHISQ=$chi_sq;
	for ($j=0;$j<$n_unc;$j++) {
            my $wave_res=$wave_unc[$j]/(1+$redshift);
            my $dust_rat=A_l(3.1,$wave_res);
            my $norm_C=0;
            my $norm_C_mass=0;
            for ($k=0;$k<$nf;$k++) {
                my $dust=10**(-0.4*$Av[$k]*$dust_rat);
                my $C=$coeffs->at($k,0);
                $norm_C=$norm_C+$C;
                $norm_C_mass=$norm_C_mass+$C*$ml[$k];
		set($coeffs_N,$k,0,$norm_C);
		set($coeffs_NM,$k,0,$norm_C);
            }
	}
	$age_min=0;
	$met_min=0;
	$Av_min=0;
	$age_min_mass=0;
	$met_min_mass=0;
	$Av_min_mass=0;

	my $norm_C=0;
	my $norm_C_mass=0;
	for ($k=0;$k<$nf;$k++) {
	    my $dust=10**(-0.4*$Av[$k]*$dust_rat);
	    my $C=$coeffs->at($k,0);
	    $norm_C=$norm_C+$C;
	    $norm_C_mass=$norm_C_mass+$C*$ml[$k];
	    set($coeffs_N,$k,0,$norm_C);
	    set($coeffs_NM,$k,0,$norm_C);
	}


	
	for ($k=0;$k<$nf;$k++) {
	    my $C=$coeffs->at($k,0);
	    if ($norm_C>0) {
		$age_min=$age_min+$C*log10($age_mod[$k])/$norm_C;
		$met_min=$met_min+$C*$met_mod[$k]/$norm_C;
		$Av_min=$Av_min+$C*$Av[$k]/$norm_C;	    
		$CN=$C/$norm_C;
	    }
	    my $C_now=$C*$med_norm;
	    if ($norm_C_mass>0) {
		$age_min_mass=$age_min_mass+$C*log10($ml[$k]*$age_mod[$k])/$norm_C_mass;
		$met_min_mass=$met_min_mass+$C*$ml[$k]*$met_mod[$k]/$norm_C_mass;
		$Av_min_mass=$Av_min_mass+$C*$ml[$k]*$Av[$k]/$norm_C_mass;
	    }

	}
	$age_min=10**($age_min);
	$age_min_mass=10**($age_min_mass);
	
    }
    my $name=$unc_file.", ";
    my $scale="1";
    
    my $pdl_age_mod=pdl(@age_mod);
    my $pdl_met_mod=pdl(@met_mod);
    my $pdl_ml=pdl(@ml);
    my $pdl_Av=pdl(@Av);
    my $pdl_model_spec_min=pdl(@model_spec_min);
    my $pdl_res=pdl(@res_spec);

#    my $pdl_model_spec_cat=zeroes($n_unc,$n_mc);

    my $out_ps_now="junk";
    my $title="X=$chi_sq Av=$Av[0] z=$redshift sigma=$sigma";
    if ($plot==1) {
	plot_results($plot,pdl(@wave_unc),pdl($pdl_flux_masked,$pdl_model_spec_min,$pdl_res),$out_ps_now,$title);
    }


#    my $end_cat=$ini_cat-1;
#    my $coeffs_cat_sec=$coeffs_cat->slice(":,0:$end_cat");

    
    return $chi_sq,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_model_spec_min,$pdl_res;
}

sub fit_ssp_lin {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;
    my $crval=$_[3];
    my $cdelt=$_[4];
    my $crpix=$_[5];
    my $nf=$_[6];
    my $n_c=$_[7];
    my $pdl_flux_c_ini=$_[8];
    my $junk=$_[9];
    my @wave_unc=@$junk;
    my $junk=$_[10];
    my @masked=@$junk;
    my $junk=$_[11];
    my @e_flux_unc=@$junk;
    my $junk=$_[12];
    my @flux_masked=@$junk;
    my $n_mc=$_[13];
    my $chi_sq_min_now=$_[14];
    my $MIN_CHISQ=$_[15];
    my $plot=$_[16];
    my $k,$j,$i,$iter;
    my $iii,$jj;
    my $iter_max=5;
    my $last_chi=1e12;

    my $coeffs=zeroes($nf,3);    
    my $coeffs_N=zeroes($nf,1);    
    my $coeffs_NM=zeroes($nf,1);    

    my $coeffs_cat=zeroes($nf+1,$n_mc*$n_mc);

    my $pdl_model_spec_cat=zeroes($n_unc,$n_mc*$n_mc);

    for ($i=0;$i<$nf;$i++) {
	if ($Av[$i]<0) {
	    $Av[$i]=0;
	}
    }

    my @wave_c;
    my @dpix_c_val;
    for ($j=0;$j<$n_c;$j++) {
	$wave_c[$j]=($crval+$cdelt*($j+1-$crpix))*(1+$redshift);
	if ($j>0) {
	    $dpix_c_val[$j]=$wave_c[$j]-$wave_c[$j-1];	
	}
	
    }

    $dpix_c_val[0]=$dpix_c_val[1];
    $dpix_c=$wave_c[1]-$wave_c[0];

    my $rsigma=$sigma/$dpix_c;

    my @name;
    my @age_mod,@met_mod;
    my @ml;
    my @flux_c;
    my $pdl_flux_c_conv;


    my @model;
    my @model_no_mask;
    
    my $y_model_end;
    my $y_model_no_mask_end;


    my @MED_NORM;

	my $age_min;
	my $met_min;
	my $Av_min;
	my $age_min_mass;
	my $met_min_mass;
	my $Av_min_mass;

    for ($iii=0;$iii<$nf;$iii++) {
	my $header="NAME".$iii;
	$name[$iii]=$pdl_flux_c_ini->hdr->{$header};
	my $name_min=$name[$iii];
	$name_min =~ s/spec_ssp_//;
	$name_min =~ s/.spec//;    
	$name_min =~ s/.dat//;
	my $AGE,$MET;
	my $age,$met;
	($AGE,$MET)=split("_",$name_min);
	if ($AGE =~ "Myr") {
	    $age=$AGE;
	    $age =~ s/Myr//;
	    $age=$age/1000;
	} else {
	    $age=$AGE;
	    $age =~ s/Gyr//;
	}
	$met=$MET;
	$met =~ s/z/0\./;	
	$age_mod[$iii]=$age;
	$met_mod[$iii]=$met;
	$header="NORM".$iii;	
	my $val_ml=$pdl_flux_c_ini->hdr->{$header};
	if ($val_ml!=0) {
	    $ml[$iii]=1/$val_ml;
	} else {
	    $ml[$iii]=1;
	}
	
	my $box=int(3*$rsigma);
	if ($box<3) {
	    $box=3;
	}
	my $kernel=zeroes(2*$box+1);
	my $norm=0;
	$flux_c[$i]=0;
	for ($j=0;$j<2*$box+1;$j++) {
	    $gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	    set($kernel,$j,$gaus);
	    $norm=$norm+$gaus;
	}
	$kernel=$kernel/$norm;
        $pdl_flux_c_conv = conv2d $pdl_flux_c_ini,$kernel;
	my $pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");
	my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c), $pdl_flux_c);
	my ($n_c_out)=$out_spec_pdl->dims;
	my @error;

	for ($i=0;$i<$n_unc;$i++) {
	    my $val=$out_spec_pdl->at($i);
	    if ($val eq "nan") {
		$val=0;
	    }
	    $model[$i][$iii]=$val*$masked[$i];
	    $model_no_mask[$i][$iii]=$val;#*$masked[$i];
	    if ($masked[$i]>0) {
		$error[$i]=0.01*abs($e_flux_unc[$i]);
	    } else {
		$error[$i]=0;
	    }

	}
    }

#    print "Av = @Av\n";


    my $pdl_model=zeroes($n_unc,$nf);
    my $pdl_model_no_mask=zeroes($n_unc,$nf);
    my $pdl_error=zeroes($n_unc);
    my $pdl_masked=pdl(@masked);
    for ($j=0;$j<$nf;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
	    my $wave_res=$wave_unc[$i]/(1+$redshift);
	    my $dust_rat=A_l(3.1,$wave_res);
	    my $dust=10**(-0.4*$Av[$j]*$dust_rat);  
	    my $val=$model[$i][$j]*$dust;
	    set($pdl_model,$i,$j,$val);
	    my $val_no_mask=$model_no_mask[$i][$j]*$dust;
	    set($pdl_model_no_mask,$i,$j,$val_no_mask);
	    my $e_val=$error[$i];
	    my $val_now=$e_flux_unc[$i];
	    if ($val_now==0) {
		$val_now=1;
	    }
	    set($pdl_error,$i,1/(abs($val_now)**2));
	}
    }
#
# We fit
#
    my $pdl_flux_masked=pdl(@flux_masked);
    my $j_mc;
    my $ini_cat=0;
    my $pdl_C_flast=ones($nf);

# Just a linear FIT, without restrictions!
    ($y_model_now, $coeffs) = my_linfit1d $pdl_flux_masked,$pdl_model,$pdl_error;

##############################
# CHISQ VALUE
	    my $chi=0;
	    my $chi2=0;
	    my $NFREE=0;
	    my @out_spec;
	    my @chi_sec;
	    for ($j=0;$j<$n_unc;$j++) {
		$out_spec[$j]=$y_model_now->at($j,0);
		$model_spec[$j]=$out_spec[$j];
		$res_spec[$j]=$flux_unc[$j]-$model_spec[$j];	    
		$model_spec_min[$j]=$model_spec[$j];
		$chi_sec[$j]=0;
		if (($flux_unc[$j]!=0)&&($out_spec[$j]!=0)) {
		    $chi=$chi+$masked[$j]*(($flux_masked[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
		    if ($have_error==0) {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs($out_spec[$j]);
		    } else {
			$chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/($e_flux_unc[$j])**2;
		    }
		    $NFREE++;
		}
	    }
	    
	    my $chi_sq=$chi;
	    if ($NFREE>0) {
		$chi_sq=($chi_sq/($NFREE))**0.5;
	    }

    $chi_sq_min_now=$chi_sq;
    $y_model_end=$y_model_now;
    $y_model_no_mask_end=$y_model_no_mask_now;

    $min_coeffs=$coeffs;

    my $chi_sq=$chi_sq_min_now;
	
    my $norm_C=0;
    my $norm_C_mass=0;
    for ($k=0;$k<$nf;$k++) {
	my $dust=10**(-0.4*$Av[$k]*$dust_rat);
	my $C=$coeffs->at($k,0);
	$norm_C=$norm_C+$C;
	    $norm_C_mass=$norm_C_mass+$C*$ml[$k];
	set($coeffs_N,$k,0,$norm_C);
	set($coeffs_NM,$k,0,$norm_C);
    }
    

	
    for ($k=0;$k<$nf;$k++) {
	my $C=$coeffs->at($k,0);
	if ($norm_C>0) {
	    $age_min=$age_min+$C*log10($age_mod[$k])/$norm_C;
	    $met_min=$met_min+$C*$met_mod[$k]/$norm_C;
	    $Av_min=$Av_min+$C*$Av[$k]/$norm_C;	    
	    $CN=$C/$norm_C;
	}
	my $C_now=$C*$med_norm;
	if ($norm_C_mass>0) {
	    $age_min_mass=$age_min_mass+$C*log10($ml[$k]*$age_mod[$k])/$norm_C_mass;
	    $met_min_mass=$met_min_mass+$C*$ml[$k]*$met_mod[$k]/$norm_C_mass;
	    $Av_min_mass=$Av_min_mass+$C*$ml[$k]*$Av[$k]/$norm_C_mass;
	}
	
	}
    $age_min=10**($age_min);
    $age_min_mass=10**($age_min_mass);
    
    my $pdl_age_mod=pdl(@age_mod);
    my $pdl_met_mod=pdl(@met_mod);
    my $pdl_ml=pdl(@ml);
    my $pdl_Av=pdl(@Av);
    my $pdl_model_spec_min=pdl(@model_spec_min);
    my $pdl_res=pdl(@res_spec);

    my $out_ps_now="junk";
    my $title="X=$chi_sq Av=$Av[0] z=$redshift sigma=$sigma";
    if ($plot==1) {
	plot_results($plot,pdl(@wave_unc),pdl($pdl_flux_masked,$pdl_model_spec_min,$pdl_res),$out_ps_now,$title);
    }

    
    return $chi_sq,$pdl_age_mod,$pdl_met_mod,$pdl_ml,$pdl_Av,$coeffs,$coeffs_N,$coeffs_NM,$pdl_model_spec_min,$pdl_res;
}


sub FIT_SINGLE {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;

    my $i,$j,$k;

#
# We renormalize the models
#


    for ($j=0;$j<$n_c_new;$j++) {
	$wave_c_new[$j]=($crval_new+$cdelt_new*($j+1-$crpix_new))*(1+$redshift);
    }
    $dpix_c=$wave_c_new[1]-$wave_c_new[0];
    $rsigma=$sigma/$dpix_c;


    my $CHI_MIN=1e12;

    $box=int(3*$rsigma);
    if ($box<3) {
	$box=3;
    }
    $kernel=zeroes(2*$box+1);
    $norm=0;
    $flux_c[$i]=0;
    for ($j=0;$j<2*$box+1;$j++) {
	$gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	set($kernel,$j,$gaus);
	$norm=$norm+$gaus;
    }
    $kernel=$kernel/$norm;
    
    $pdl_flux_c_conv = conv2d $pdl_flux_c_ini_new,$kernel;
    

    for ($iii=0;$iii<$nf_new;$iii++) {
	$header="NAME".$iii;
	$name[$iii]=$pdl_flux_c_ini_new->hdr->{$header};
	$name_min=$name[$iii];
	$name_min =~ s/spec_ssp_//;
	$name_min =~ s/.spec//;
	($AGE,$MET)=split("_",$name_min);
	if ($AGE =~ "Myr") {
	    $age=$AGE;
	    $age =~ s/Myr//;
	    $age=$age/1000;
	} else {
	    $age=$AGE;
	    $age =~ s/Gyr//;
	}
	$met=$MET;
	$met =~ s/z/0\./;

	$age_mod[$iii]=$age;
	$met_mod[$iii]=$met;
#****** CONVOLUTION
	

	$pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");

	my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c_new), $pdl_flux_c);
	($n_c_out)=$out_spec_pdl->dims;
	

	
	
	for ($i=0;$i<$n_unc;$i++) {
	    if ($masked[$i]>0) {
		$error[$i]=0.01/abs($e_flux_unc[$i]);
	    } else {
		$error[$i]=0;
	    }
	    $val=$out_spec_pdl->at($i);
	    $model[$i][$iii]=$val;#*$masked[$i];
#	    if ($NITER==0) {
		for ($j=0;$j<$nline;$j++) {
		    if (abs($wave_unc[$i]-$w_eline[$j]*(1+$redshift))<(3*$rsigma)) {
#		    $error[$i]=(30/(1+$INTER))*$e_flux_unc[$i];
		    $error[$i]=0.1/abs($e_flux_unc[$i]);
#			$error[$i]=10*$e_flux_unc[$i];
		    }
		}
	    }
#	}
	
    }



    $pdl_model=zeroes($n_unc,$nf_new);
    $pdl_error=zeroes($n_unc,$nf_new);
    my $pdl_masked=pdl(@masked);
    for ($j=0;$j<$nf_new;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
#	    if ($j==0) {
		$wave_res=$wave_unc[$i]/(1+$redshift);
		$dust_rat=A_l(3.1,$wave_res);
		$dust=10**(-0.4*$Av[$j]*$dust_rat);  
		$a_dust[$i]=$dust;
#	    }
#	    print "DUST = $val $model[$i][$j] $a_dust[$i]\n";
	    $val=$model[$i][$j]*$a_dust[$i];
	    set($pdl_model,$i,$j,$val);
	    $e_val=$error[$i];
	    set($pdl_error,$i,$j,1);
	}
    }
    $pdl_flux_masked=pdl(@flux_masked);


#    ($yfit, $coeffs) = my_linfit1d $pdl_flux_masked,$pdl_model,$pdl_error;
    #
    # We normalize the models
    #
 

#    my $nx1=int(($wave_norm-$w_wave_norm-$crval3)/$cdelt3);
#    my $nx2=int(($wave_norm+$w_wave_norm-$crval3)/$cdelt3);
#    my $pdl_sec_cen=$pdl_flux_masked->slice("$nx1:$nx2");
#    my $SUM=sum($pdl_sec_cen);
    
    my $pdl_chi=zeroes($nf_new);
    my $pdl_e=pdl(@error);
#    print "$pdl_e\n";
    $J_MIN=-1;
   for ($j=0;$j<$nf_new;$j++) {
       my $pdl_model_tmp=$pdl_model->slice(":,$j");
#       my $pdl_sec_cen=$pdl_model_tmp->slice("$nx1:$nx2");
#       my $SUM_NOW=sum($pdl_sec_cen);
#       my $NORM=$SUM/$SUM_NOW;
#       $pdl_model_tmp=$pdl_model_tmp*$NORM;
       ($yfit, $coeffs) = my_linfit1d $pdl_flux_masked,$pdl_model_tmp,$pdl_e;


       my $pdl_chi_now=$pdl_masked*(($pdl_flux_masked-$yfit)**2)*$pdl_e;
       my $chi_sq=sum($pdl_chi_now);
       $chi_sq=($chi_sq/($n_unc-1))**0.5;
       set($pdl_chi,$j,$chi_sq);
#       print "$j $CHI_MIN $chi_sq\n";

       if ($CHI_MIN>$chi_sq) {
	   $CHI_MIN=$chi_sq;
	   $J_MIN=$j;
	   $NORM_J=$NORM;


#	   @out_spec=list($pdl_model_tmp);
#	   @model_spec_min=list($pdl_model_tmp);
	   @out_spec=list($yfit);
	   @model_spec_min=list($yfit);
	   my $pdl_res=pdl(@flux_unc)-$pdl_model_tmp;
	   @res_spec=list($pdl_res);
	   $JJ=0;
	   for ($jj=0;$jj<$n_unc;$jj++) {
	       if ($masked[$jj]>0) {
		   $res_clean[$JJ]=$res_spec[$jj];
		   $flux_clean[$JJ]=$flux_unc[$jj];
		   $JJ++;
	       }
	   }




	if ($plot==1) {

	    pgbegin(0,$dev_plot_single,1,1);
	    pgsfs(1.2);
	    pgscf(2);             # Set character font
	    pgslw(2);             # Set line width
	    pgsch(1.2);           # Set character height
	    pgenv($wave_unc[0],$wave_unc[$n_unc-1],$y_min,$y_max,0,0);
	    pgsch(1.2);           # Set character height
	    pglabel("Wavelength","Counts","SSP X=$CHI_MIN T=$age_mod[$J_MIN] Z=$met_mod[$J_MIN] z=$redshift sigma=$sigma Av=@Av");
	    pgsci(1);
	    pgline($n_unc,\@wave_unc,\@flux_unc);    
	    pgsci(2);
	    pgline($n_unc,\@wave_unc,\@out_spec);    
	    pgsci(3);
	    pgline($n_unc,\@wave_unc,\@flux_masked);    
	    pgsci(1);
	    pgclose;
	    pgend;
#    print "Press Enter"; <stdin>;
	}

       }

   }
#    <stdin>;
#    $pdl_model_tmp=$pdl_model->slice(":,$J_MIN");
#    $pdl_model_tmp=$pdl_model_tmp*$NORM_J;



#    print "SSP CHI=$CHI_MIN $age_mod[$J_MIN] $met_mod[$J_MIN]\n";

    if ($CHI_MIN<$MIN_CHISQ) {    
	$MIN_CHISQ=$CHI_MIN;
	@Av_MIN=@Av;
	$age_min=$age_mod[$J_MIN];
	$met_min=$met_mod[$J_MIN];

    }
    return($CHI_MIN);
}



################ FIT_NEW #########################


sub FIT_Z_SIGMA {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;
    my $k,$j,$i,$iter;

    for ($j=0;$j<$n_c;$j++) {
	$wave_c[$j]=($crval+$cdelt*($j+1-$crpix))*(1+$redshift);
	if ($j>0) {
	    $dpix_c_val[$j]=$wave_c[$j]-$wave_c[$j-1];	
	}
	
    }
    $dpix_c_val[0]=$dpix_c_val[1];
    $dpix_c=$wave_c[1]-$wave_c[0];
    $rsigma=$sigma/$dpix_c;

# CONVOLUTION
    $box=int(3*$rsigma);
    if ($box<3) {
	$box=3;
    }
    $kernel=zeroes(2*$box+1);
    $norm=0;
    for ($j=0;$j<2*$box+1;$j++) {
	$gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	set($kernel,$j,$gaus);
	$norm=$norm+$gaus;
    }
    $kernel=$kernel/$norm;

    $pdl_flux_c_conv = conv2d $pdl_flux_c_ini,$kernel;
# END CONVOLUTION

    $pdl_dust=zeroes($n_unc,$nf);
    for ($j=0;$j<$nf;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
	    $wave_res=$wave_unc[$i]/(1+$redshift);
	    $dust_rat=A_l(3.1,$wave_res);
	    if ($Av[$j]!=0) {
		$dust=10**(-0.4*$Av[$j]*$dust_rat);  
		set($pdl_dust,$i,$j,$dust);
	    } else {
		set($pdl_dust,$i,$j,1);
	    }
	}
    }


    my $model_now=zeroes($n_unc);    
    for ($iii=0;$iii<$nf;$iii++) {    
	$pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");
	my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c), $pdl_flux_c);
	$C=$coeffs->at($iii,0);
	my $dust_spec=$out_spec_pdl*$pdl_dust->slice(",$iii");
	$model_now=$model_now+$C*$dust_spec;
    }

    $chi=0;
    $chi2=0;
    $NFREE=0;
    for ($j=0;$j<$n_unc;$j++) {
	if (($wave_unc[$j]>$MIN_W)&&($wave_unc[$j]<$MAX_W)) {
	    $out_spec[$j]=$model_now->at($j,0);
	    if (($flux_unc[$j]!=0)&&($out_spec[$j]!=0)) {
		$chi_j[$j]=$masked[$j]*(($flux_masked[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
		$chi=$chi+$chi_j[$j];
		$NFREE++;
	    }
	}
    }

   $chi_sq=$chi;
    if ($NFREE>0) {
	$chi_sq=($chi_sq/($NFREE))**0.5;
    }
    return $chi_sq;
}





sub FIT_simple {
    my $redshift=$_[0];
    my $sigma=$_[1];
    my $Av_NOW=$_[2];
    my @Av=@$Av_NOW;
#    print "AV FIT = '@AV' '$Av_NOW' '$_'\n";
    my $k,$j,$i,$iter;

    for ($i=0;$i<$nf;$i++) {
	if ($Av[$i]<0) {
	    $Av[$i]=0;
	}
    }


    for ($j=0;$j<$n_c;$j++) {
	$wave_c[$j]=($crval+$cdelt*($j+1-$crpix))*(1+$redshift);
	if ($j>0) {
	    $dpix_c_val[$j]=$wave_c[$j]-$wave_c[$j-1];	
#	    print "$j $dpix_c_val[$j] $crval $cdelt\n";
	}
	
    }
    $dpix_c_val[0]=$dpix_c_val[1];
$dpix_c=$wave_c[1]-$wave_c[0];

#$dpix_c=median($wave_c[1]-$wave_c[0];

#($dpix_c,$dpix_max)=minmax(@dpix_c_val);
#$dpix_c=median(@dpix_c_val);
#print "DPIX= $dpix_c\n"; exit;

$rsigma=$sigma/$dpix_c;

#    print "$sigma $dpix_c\n";

for ($iii=0;$iii<$nf;$iii++) {
    $header="NAME".$iii;
    $name[$iii]=$pdl_flux_c_ini->hdr->{$header};
    $name_min=$name[$iii];
    $name_min =~ s/spec_ssp_//;
    $name_min =~ s/.spec//;    
    $name_min =~ s/.dat//;
    ($AGE,$MET)=split("_",$name_min);
    if ($AGE =~ "Myr") {
	$age=$AGE;
	$age =~ s/Myr//;
	$age=$age/1000;
    } else {
	$age=$AGE;
	$age =~ s/Gyr//;
    }
    $met=$MET;
    $met =~ s/z/0\./;

    $age_mod[$iii]=$age;
    $met_mod[$iii]=$met;
    $header="NORM".$iii;
#    $ml[$iii]=$pdl_flux_c_ini->hdr->{$header};

 $val_ml=$pdl_flux_c_ini->hdr->{$header};
    if ($val_ml!=0) {
	$ml[$iii]=1/$val_ml;
    } else {
	$ml[$iii]=1;
    }




#    print "$name[$iii] $age_mod[$iii] $met_mod[$iii]\n";
    
    
    $box=int(3*$rsigma);
    if ($box<3) {
	$box=3;
    }
    $kernel=zeroes(2*$box+1);
    $norm=0;
    $flux_c[$i]=0;
    for ($j=0;$j<2*$box+1;$j++) {
	$gaus=exp(-0.5*((($j-$box)/$rsigma)**2));    
	set($kernel,$j,$gaus);
	$norm=$norm+$gaus;
    }
    $kernel=$kernel/$norm;
    

    $pdl_flux_c_conv = conv2d $pdl_flux_c_ini,$kernel;



    $pdl_flux_c = $pdl_flux_c_conv->slice(",$iii");


     my $out_spec_pdl = interpol(pdl(@wave_unc), pdl(@wave_c), $pdl_flux_c);
    ($n_c_out)=$out_spec_pdl->dims;
#
# We store the result!
#
    for ($i=0;$i<$n_unc;$i++) {
#	print "$i/$n_c\n";
	$val=$out_spec_pdl->at($i);
	if ($val eq "nan") {
	    $val=0;
	}
	$model[$i][$iii]=$val*$masked[$i];
	$model_no_mask[$i][$iii]=$val;#*$masked[$i];

	if ($masked[$i]>0) {
	    $error[$i]=0.01*abs($e_flux_unc[$i]);
#	    $error[$i]=abs($e_flux_unc[$i]);
	} else {
	    $error[$i]=0;
	}


	
#	if ($NITER==0) {
	    for ($j=0;$j<$nline;$j++) {
		if (abs($wave_unc[$i]-$w_eline[$j]*(1+$redshift))<(3*$rsigma)) {
#		    $error[$i]=0.1*$e_flux_unc[$i];
#		    $error[$i]=30;#*$e_flux_unc[$i];
		}
	    }
#	}

#		    print "$i $error[$i]\n";

#	print "$i $wave_unc[$i] $error[$i] $e_flux_unc[$i]\n";

    }
 #   print "$iii/$nf files\n";
}

    $pdl_model=zeroes($n_unc,$nf);
#    $pdl_error=zeroes($n_unc,$nf);
    $pdl_error=zeroes($n_unc);
    my $pdl_masked=pdl(@masked);
    for ($j=0;$j<$nf;$j++) {
	for ($i=0;$i<$n_unc;$i++) {
	    $wave_res=$wave_unc[$i]/(1+$redshift);
	    $dust_rat=A_l(3.1,$wave_res);
	    $dust=10**(-0.4*$Av[$j]*$dust_rat);  
	    $val=$model[$i][$j]*$dust;
	    set($pdl_model,$i,$j,$val);
	    $e_val=$error[$i];
	    $val_now=$e_flux_unc[$i];
	    if ($val_now==0) {
		$val_now=1;
	    }
#	    set($pdl_error,$i,$j,1/(abs($val_now)**2));
	    set($pdl_error,$i,1/(abs($val_now)**2));
	}
    }
#
# We fit
#
#    print "We fit with Av=$Av ";
    $pdl_flux_masked=pdl(@flux_masked);
    ($yfit, $coeffs) = my_linfit1d $pdl_flux_masked,$pdl_model,$pdl_error;


#
# We remove the models that are negative
#
    my $nf_new=0;
    $nf_neg=0;
    for ($k=0;$k<$nf;$k++) {
	$C=$coeffs->at($k,0);
	if ($C>0) {
	    $nf_new++;
	} else {
	    $nf_neg++;
	}
    }

#    print "\n";

   
    if ($nf_new>0) {
	while ($nf_neg>0) {
#	    print "PASO $nf_new $nf_neg \n";
#	    if ($nf_new>0) {
		my $pdl_model_new=zeroes($n_unc,$nf_new);
#		my $pdl_error_new=zeroes($n_unc,$nf_new);
		my $pdl_error_new=zeroes($n_unc);
#	    }
	    $nf_i=0;
#	    print "C= ";
	    for ($k=0;$k<$nf;$k++) {
		$C=$coeffs->at($k,0);
#		print " $C ";
#		print "$C $k PASO\n";
		if ($C>0) {
		    for ($i=0;$i<$n_unc;$i++) {
			$val=$pdl_model->at($i,$k);
			set($pdl_model_new,$i,$nf_i,$val);
#			print "$i PASO\n";
#			$e_val=$pdl_error->at($i,$k);
			$e_val=$pdl_error->at($i);
#			set($pdl_error_new,$i,$nf_i,$e_val);		    
			set($pdl_error_new,$i,$e_val);		    
		    }
		    $MOD[$nf_i]=$k;
		    $nf_i++;
		} else {
		    set($coeffs,$k,0,0);
		}
	    }
#	    print "\n";
#	    print "PASO\n";
	    ($yfit, $coeffs_new) = my_linfit1d $pdl_flux_masked,$pdl_model_new,$pdl_error_new;	
#	    print "FIT $coeffs_new\n";
	    
	    $nf_i=0;
	    $nf_neg=0;
	    $nf_new=0;
#	    print "C= ";
	    for ($k=0;$k<$nf;$k++) {
		$C=$coeffs->at($k,0);
#		print "$C ";
		if ($C>0) {
		    $val=$coeffs_new->at($nf_i,0);
		    $nf_i++;
		    if ($val>0) {
			set($coeffs,$k,0,$val);
			$nf_new++;
		    } else {
			set($coeffs,$k,0,0);
			$nf_neg++;
		    }
#		    print "$val , ";
		}
	    }
#	    print "\n";
		if ($nf_new==0) {
		    $nf_neg=0;
		}
	}
#	    print "$nf_new $nf_neg ";
#	    print "C= ";
	    for ($k=0;$k<$nf;$k++) {
		$C=$coeffs->at($k,0);
#		print " $C ";
	    }
#	print "\n";

    } else {
	$nf_new=$nf;
	$pdl_model_new=$pdl_model;
	$pdl_error_new=$pdl_error;
    }



#    print "PASO\n";

#
# We determine the chi**2
#
    @J=$yfit->dims;
    @K=$coeffs->dims;
#    print "DONE ( $coeffs)\n"; <stdin>;
    $chi=0;
    $chi2=0;
    $NFREE=0;
    for ($j=0;$j<$n_unc;$j++) {
	$out_spec[$j]=($yfit->at($j,0));#
	$chi_sec[$j]=0;
	    if (($flux_unc[$j]!=0)&&($out_spec[$j]!=0)) {
		#print "$j $flux_unc[$j]-$out_spec[$j]\n";
#0000000000000000000000000000000
#		$chi=$chi+$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
		$chi=$chi+$masked[$j]*(($flux_masked[$j]-$out_spec[$j])**2)/abs(1.5*$out_spec[$j]);
#		$chi2=$chi2+$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs($e_flux_unc[$j]);
		if ($have_error==0) {
		    $chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/abs($out_spec[$j]);
		} else {
		    $chi_sec[$j]=$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)/($e_flux_unc[$j])**2;
		}
#		$chi=$chi+$masked[$j]*(($flux_unc[$j]-$out_spec[$j])**2)*$error[$j];
		$NFREE++;
	    }
    }

    $chi_sq=$chi;
    if ($NFREE>0) {
	$chi_sq=($chi_sq/($NFREE))**0.5;
    }
#    $chi2=($chi2/($NFREE))**0.5;

#    $chi_sq=(($chi/($n_unc-$nf))**0.5);
#    $chi_sq=(($chi/($n_unc-1))**0.5);
    if ($chi_sq<$MIN_CHISQ) {
#    print "PASO $chi_sq $MIN_CHISQ\n";
	$MIN_CHISQ=$chi_sq;
	@Av_MIN=@Av;
	for ($j=0;$j<$n_unc;$j++) {
	    $model_spec[$j]=0;
	    $wave_res=$wave_unc[$j]/(1+$redshift);
	    $dust_rat=A_l(3.1,$wave_res);
	    $norm_C=0;
	    $norm_C_mass=0;
	    for ($k=0;$k<$nf;$k++) {
		$dust=10**(-0.4*$Av[$k]*$dust_rat);  
		$C=$coeffs->at($k,0);
		$norm_C=$norm_C+$C;
		$norm_C_mass=$norm_C_mass+$C*$ml[$k];
		$model_spec[$j]=$model_spec[$j]+$C*$model_no_mask[$j][$k]*$dust;
		$model_spec_min[$j]=$model_spec[$j];
	    }
	    $res_spec[$j]=$flux_unc[$j]-$model_spec[$j];	    
#	    print "$wave_unc[$j] $flux_unc[$j] $model_spec[$j] $res_spec[$j]\n";
	}
	$age_min=0;
	$met_min=0;
	$Av_min=0;
	$age_min_mass=0;
	$met_min_mass=0;
	$Av_min_mass=0;
	open(C,">coeffs.out");
	print C "ID   AGE     MET    COEFF   Norm.Coeff  M/L   AV\n";
	print "---------------------------------------------\n";
	print "ID AGE     MET    COEFF   Norm.Coeff   M/L    AV\n";
	print "---------------------------------------------\n";
	for ($k=0;$k<$nf;$k++) {
	    $C=$coeffs->at($k,0);
	    if ($norm_C>0) {
		$age_min=$age_min+$C*log10($age_mod[$k])/$norm_C;
		$met_min=$met_min+$C*$met_mod[$k]/$norm_C;
		$Av_min=$Av_min+$C*$Av[$k]/$norm_C;	    
		$CN=$C/$norm_C;
	    }
#	    if ($C==0) {
#		$d_Av[$k]=0;
#	    }
#	    print C "$k $age_mod[$k] $met_mod[$k] $C $CN $ml[$k] $Av[$k]\n";
	    printf(C "%-2d %-7.4f %-7.4f %-7.4f %-7.4f %-10.2f %-4.2f\n",$k,$age_mod[$k],$met_mod[$k],$C,$CN,$ml[$k],$Av[$k]);
	    if ($C>0) {
		printf("%-2d %-7.4f %-7.4f %-7.4f %-7.4f %-10.2f %-4.2f\n",$k,$age_mod[$k],$met_mod[$k],$C,$CN,$ml[$k],$Av[$k]);
	    }
	    if ($norm_C_mass>0) {
		$age_min_mass=$age_min_mass+$C*log10($ml[$k]*$age_mod[$k])/$norm_C_mass;
		$met_min_mass=$met_min_mass+$C*$ml[$k]*$met_mod[$k]/$norm_C_mass;
		$Av_min_mass=$Av_min_mass+$C*$ml[$k]*$Av[$k]/$norm_C_mass;
	    }
#		print "COEFF=$k $C $norm_C_mass $age_min $age_mod[$k] $met_min $met_mod[$k] $ml[$k] $age_min_mass $met_min_mass\n";

	}
	print "---------------------------------------------\n";
	close(C);
	$age_min=10**($age_min);
	$age_min_mass=10**($age_min_mass);
	
    }
    $name=$unc_file.", ";
    $scale="1";
#    print "*****PASO 1***** PLOT = $plot, N_UNC = $n_unc, $nc\n";
    
    if ($plot==1) {

	pgbegin(0,$dev_plot,1,1);
	pgsfs(1.2);
	pgscf(2);             # Set character font
	pgslw(2);             # Set line width
	pgsch(1.2);           # Set character height
	pgenv($min_wave,$max_wave,$y_min,$y_max,0,0);
	pgsch(0.9);           # Set character height
#	pglabel("Wavelength","Counts","X=$chi_sq ($chi_joint) T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) Av=@Av z=$redshift sigma=$sigma");
	my $mean_Av=mean(@Av);
	pglabel("Wavelength","Counts","X=$chi_sq ($chi_joint) T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) Av=$mean_Av z=$redshift sigma=$sigma");
#	pglabel("Wavelength","Counts","X=$chi_sq ($chi_joint) T=$age_min ($age_min_mass) Z=$met_min ($met_min_mass) z=$redshift sigma=$sigma");
	pgsch(0.5);
	pgsci(1);
#	print "**** PASO NOW ****\n";


	pgline($n_unc,\@wave_unc,\@flux_unc);    
	pgsci(2);
	pgline($n_unc,\@wave_unc,\@out_spec);    
	pgsci(8);
	pgline($n_unc,\@wave_unc,\@model_spec_min);    
	pgsci(5);
	pgline($n_unc,\@wave_unc,\@res_spec);    

	if ($nc>0) {
	    pgsci(6);
	    pgline($nc,\@wave_clean,\@res_clean);    
	}
	pgsci(1);
	pgsci(13);
	pgsls(2);
	$j_plot=0;



	for ($j=0;$j<$nf;$j++) {
	    $C=$coeffs->at($j,0);
	    if ($C>0) {
		pgsci(13);
		my $model_now=$pdl_model->slice(",($j)");

		$model_now=$C*$model_now;
		@a_model_now=list($model_now);
		pgline($n_unc,\@wave_unc,\@a_model_now);    

	    }
	    if ($norm_C != 0) {
		$CN=$C/$norm_C;
	    } else {
		$CN=$C;
	    }
	    if ($C>0) {
		$pos_x=$min_wave*1.02;
		$pos_y=$y_max-0.05*($j_plot+1)*($y_max-$y_min);
		pgsch(0.8);           # Set character height
		if ($C>0) {
		    pgsci(1);
		} else {
		pgsci(15);
		}
		$CN=apr($CN);
		pgptxt($pos_x,$pos_y,0,0,"$CN $age_mod[$j] $met_mod[$j]");
		$j_plot++;
#	    print "$pos_x $pos_y\n"; <stdin>;
	    }
		pgsch(1.2);           # Set character height

	}
	pgsls(1);


	pgsci(1);


	pgclose;
	pgend;
#    print "Press Enter"; <stdin>;
    }


#
#
#





    return $chi_sq;
}


sub get_time {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $time="# TIME ".$sec." ".$min." ".$hour." ".$mday." ".$mon." ".$year." ".$wday." ".$yday." ".$isdst;
    return($time);
}


sub fit_elines_rnd() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_mc=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
#    my $SCALE_INI=3;
    my $SCALE=$SCALE_INI;
#    print "N.MOD = $n_mod, N.MC=$n_mc\n";

    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }
#    print "$nx,$ny\n";
#    my $pdl_model=zeroes($nx,$ny);
#    my $pdl_masked=ones($nx);

    my $i;
    #
    #

    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }


    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
#    print "CHI_sq [GUESS]=$chi_sq\n";  
	$chi_sq_ini=$chi_sq;
	@a_out=copy_a($n_mod,\@a);




    #############################
    # TEST OUTPUT!
    #
    open(OUT,">test_fit_elines_rnd.spec");
    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
	my $f_mod=$pdl_model_0->at($i)*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont_0->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6450 6800 &";
    system($call);
#    print "###################################\n";
#    <stdin>;
    #
    # TEST OUTOUT
    ###############################

    my $i,$j,$ii;
    my @a_now;

#    srand(localtime);
	$pdl_rnd=random($n_mod*9*$n_mc);

    for ($ii=0;$ii<$n_mc;$ii++) {
	for ($i=0;$i<$n_mod;$i++) {
	    for ($j=0;$j<9;$j++) {
		$rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
		if ($ia[$i][$j]==1) {
		    if ($link[$i][$j]==-1) {
			$a_now[$i][$j]=$a0[$i][$j]+($a1[$i][$j]-$a0[$i][$j])*$rnd;	
		    } else {
			$k=$link[$i][$j]-1;
			$method=$a1[$i][$j];
			if ($method==0) {
			    $a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
			} else {
			    $a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
			}			    
		    }
		} else {
		    $a_now[$i][$j]=$a[$i][$j];
		}	    
	    }
	}

    # MC test
    $pdl_model_0=zeroes($nx);
    $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }
#    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);

    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
	if ($chi_sq<$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;


    #############################
    # TEST OUTPUT!
    #
    open(OUT,">test_fit_elines_rnd.spec");
    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
		my $f_mod=$pdl_model_0->at($i)*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont_0->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
#    print "iter= $ii/$n_mc  CHI_sq=$chi_sq / $chi_sq_ini SCALE = $SCALE\n";	    
#	    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6300 7000 ";
    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6450 6800 &";
    system($call);
    #
    # TEST OUTOUT
	    ###############################


	    my $iii;
	    for ($iii=0;$iii<$n_mod-1;$iii++) {
		if (($type[$iii] eq "eline") and ($link[$iii][1]==-1)) {
		    $a0[$iii][1]=$a_out[$iii][1]/(1+1/$SCALE);
		    $a1[$iii][1]=$a_out[$iii][1]*(1+$SCALE);
		}
	    }


	    $SCALE=0.75*$SCALE;
	    if ($SCALE>0.001) {
		$ii=0;
	    } else {
		$ii=$n_mc;
	    }








	} else {

	#    $SCALE=1.1*$SCALE;

	}
#else {	    
#	    if (($def==1)&&($chi_sq>2*$chi_sq_ini)) {
#	    if ($def==3) {
#
#		}
#	    }
#	}

	   
    }
    $chi_sq_now=$chi_sq_ini;





    return @a_out;
}




sub get_initial_guess() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $i_now=$_[2];
    my $junk=$_[3];
    my @type=@$junk;
    my $junk=$_[4];
    my @a=@$junk;
    my @dims=$pdl_wave->dims();
    my $nx=$dims[0];
    my $i=0;
#    my $pdl_out=zeroes($nx);
    my $j=0;
    my $guess;

    if ($type[$i_now] eq "eline") {
	$guess=$a[$i_now][1];
	$min_delta_wave=1e12;
	for ($i=0;$i<$nx;$i++) {
	    my $w=$pdl_wave->at($i);	    
	    my $speed_of_light=299792.458;
	    my $factor=(1+$a[$i_now][3]/$speed_of_light+$a[$i_now][5]);
	    my $delta_wave=abs($w-$a[$i_now][0]*$factor);
	    if ($min_delta_wave>$delta_wave) {
		$min_delta_wave=$delta_wave;
		$guess=$pdl_flux->at($i);
		$guess=$guess*($a[$i_now][2]*((2*3.1416)**0.5));
	    }
	}
    }

    if ($type[$i_now] eq "poly1d") {
	$guess=$a[$i_now][0];
	@stats=stats($pdl_flux);
	$guess=$stats[2];
    }
    return $guess;
}



sub fit_elines_grad_rnd() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_mc=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out=copy_a($n_mod,\@a);

    my $SCALE=$SCALE_INI;
   

    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;

    my $pdl_model=zeroes($nx);
    my $pdl_model_cont=zeroes($nx);
    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;


    # We start the fitting loop!
    $pdl_rnd=grandom($n_mod*9*$n_mc);
    $pdl_rnd_lin=random($n_mod*9*$n_mc);
    my $ii=0;
    while ($ii<$n_mc) {

	#
	# We change slighly the parameters
	#
	for ($i=0;$i<$n_mod;$i++) {
	    for ($j=0;$j<9;$j++) {
		$rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
		$rnd_lin=$pdl_rnd_lin->at(($j+(9*$i))*$ii);
		if ($ia[$i][$j]==1) {
		    if ($link[$i][$j]==-1) {
			if ($j==3) {
			    $a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$rnd*($a1[$i][$j]-$a0[$i][$j])/5;
			} else {
			    $a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$a_out[$i][$j]*$rnd;
			}
			if ($a_now[$i][$j]<$a0[$i][$j]) {
			    $a_now[$i][$j]=$a0[$i][$j];
			}
			if ($a_now[$i][$j]>$a1[$i][$j]) {
			    $a_now[$i][$j]=$a1[$i][$j];
			}
		    } else {
			$k=$link[$i][$j]-1;
			$method=$a1[$i][$j];
			if ($method==0) {
			    $a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
			} else {
			    $a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
			}			    
		    }
		} else {
		    $a_now[$i][$j]=$a_out[$i][$j];
		}	    
#		print "$a_now[$i][$j] $ia[$i][$j] $a0[$i][$j] $a1[$i][$j] $link[$i][$j]\n";
	    }
	}
	
	# MC test
	$pdl_model_0=zeroes($nx);
	$pdl_model_cont_0=zeroes($nx);
	for ($i=0;$i<$n_mod;$i++) {
	    my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	    $pdl_model_0=$pdl_model_0+$pdl_tmp;
	    if ($type[$i] eq "poly1d") {
		$pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	    }
	}
#	$pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);
	my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
	my $chi_sq=sum($pdl_chi_now);
	$chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
	if ($chi_sq<=$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;
	    $pdl_model=$pdl_model_0;
	    $pdl_model_cont=$pdl_model_cont_0;
#	    $ii=0;
	} else {
	    if ((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal)) {
		$ii=$n_mc;
	    }
	}
	$ii++;
    }
    $chi_sq_now=$chi_sq_ini;
    my $pdl_a_out=pdl(@a_out);
    return ($chi_sq_now,$pdl_a_out,$pdl_model,$pdl_model_cont);
}


sub fit_elines_grad_rnd_NEW() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_mc=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];    
    my $SCALE=$SCALE_INI;


    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }
#    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;

    my $pdl_rnd=grandom($n_mod*9*$n_mc);
    my $pdl_rnd_lin=random($n_mod*9*$n_mc);

    # We start the fitting loop!
    my $ii=0;
    while ($ii<$n_mc) {
	my $pdl_chi_now;
	my $chi_sq;
	my $II;
	my @childs;
	my $pid;


	my $junk=fit_elines_rnd_single($pdl_wave,$pdl_flux,$pdl_e_flux,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out,\@ia,\@a0,\@a1,\@link,$n_mc,$pdl_masked,$def,$scale_ini,$pdl_rnd,$pdl_rnd_lin,$ii);
	my @a_now=@$junk;
	$chi_sq=$a_now[$n_mod][0];

#	$chi_sq=$chi_single;

#	print "$ii,$n_mc,$chi_sq,$chi_sq_ini\n";

	
	if ($chi_sq<$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;
	    $pdl_model=$pdl_model_0;
	    $pdl_model_cont=$pdl_model_cont_0;
	} else {
	    if ((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal)) {
		$ii=$n_mc;
	    }
	}

	$ii++;
    }
    $chi_sq_now=$chi_sq_ini;
    return @a_out;
}



sub fit_elines_grad_rnd_fork() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_mc=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my $NC=$_[16];
    my @a_out;    
    my $SCALE=$SCALE_INI;

    my @a_fork;
    
    my $i; 


    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }
#    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;

    my $pdl_rnd=grandom($n_mod*9*$n_mc);
    my $pdl_rnd_lin=random($n_mod*9*$n_mc);

    # We start the fitting loop!
    my $ii=0;
    my @a_now_final;
    my @a_now_single;

    while ($ii<$n_mc) {
	my $pdl_chi_now;
	my $chi_sq;
	my $II;
	my @childs;
	my $pid;

# $forks[$i] = async { return sum($i,$sum); };

	for ($II=0;$II<$NC;$II++) {
	    $a_now_single[$II]= async {     my $pdl_rnd=grandom($n_mod*9*$n_mc);
					    my $pdl_rnd_lin=random($n_mod*9*$n_mc);
					    fit_elines_rnd_single($pdl_wave,$pdl_flux,$pdl_e_flux,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out,\@ia,\@a0,\@a1,\@link,$n_mc,$pdl_masked,$def,$scale_ini,$pdl_rnd,$pdl_rnd_lin,$ii);
	    }
	}
	

	for ($II=0;$II<$NC;$II++) {
	    my $junk = $a_now_single[$II]->join();
	    @a_now=@$junk;
	    $chi_sq=$a_now[$n_mod][0];
#	    print "$ii,$II $chi_sq\n";
	    if ($chi_sq<$chi_sq_ini) {
		@a_out=copy_a($n_mod,\@a_now);
		$chi_sq_ini=$chi_sq;
		$pdl_model=$pdl_model_0;
		$pdl_model_cont=$pdl_model_cont_0;
	    } else {
		if ((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal)) {
		    $ii=$n_mc;
		}
	    }
	    
	}
	$ii=$ii+$NC;
	$chi_sq_now=$chi_sq_ini;
    }
    return @a_out;
}



sub fit_elines_rnd_single() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_mc=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
    my $SCALE=$SCALE_INI;
    my $pdl_rnd=$_[16];
    my $pdl_rnd_lin=$_[17];
    my $ii=$_[18];
    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }
#    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;


    # We start the fitting loop!

    my $pdl_chi_now;
    my $chi_sq;
    my $II;
    my @childs;
    my $pid;


#    print_a_final($n_mod,\@a_out,\@type,$chi_sq_now);

    
	# We change slighly the parameters
    #
    for ($i=0;$i<$n_mod;$i++) {
	for ($j=0;$j<9;$j++) {
	    $rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
	    $rnd_lin=$pdl_rnd_lin->at(($j+(9*$i))*$ii);
	    if ($ia[$i][$j]==1) {
		if ($link[$i][$j]==-1) {
		    if ($j==3) {
			$a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$rnd*($a1[$i][$j]-$a0[$i][$j])/5;
		    } else {
			$a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$a_out[$i][$j]*$rnd;
		    }
		    if ($a_now[$i][$j]<$a0[$i][$j]) {
			$a_now[$i][$j]=$a0[$i][$j];
		    }
		    if ($a_now[$i][$j]>$a1[$i][$j]) {
			$a_now[$i][$j]=$a1[$i][$j];
		    }
		} else {
		    $k=$link[$i][$j]-1;
		    $method=$a1[$i][$j];
		    if ($method==0) {
			$a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
		    } else {
			$a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
		    }			    
		}
	    } else {
		$a_now[$i][$j]=$a_out[$i][$j];
	    }	    
#	    print "$i,$j $a_now[$i][$j]\n";
	}
    }
    
    # MC test
    $pdl_model_0=zeroes($nx);
    $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
	if ($type[$i] eq "poly1d") {
	    $pdl_model_cont_0=$pdl_model_cont_0+$pdl_tmp;
	}
    }
#    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);
    $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_single=$chi_sq;
    $a_now[$n_mod][0]=$chi_sq;
#    print "@a_now\n";
    #$chi_sq_now=$chi_sq;
    #print "$chi_sq_now\n";

    return \@a_now;
}

sub plot_results {
    my $plot=$_[0];
    my $pdl_wave=$_[1];
    my $pdl_output=$_[2];
    my $output_name=$_[3];
    my $title=$_[4];
    my $min_wave,$max_wave;
    my $dev_plot;
    my $y_min,$y_max;
    my @stats;
    my $nx,$ny;
    my @wave_now=list($pdl_wave);
    my $i,$j;

    if ($plot>0) {
	$dev_plot="/null";
	if ($plot==1) {
	    $dev_plot="/xs";
	} else {
	    if ($plot==3) {
		$dev_plot=$output_name.".png/PNG";
	    } else {
		$dev_plot=$output_name.".ps/CPS";
	    }
	}
#	print "PLOT=$plot,DEV=$dev_plot\n";
	@stats=stats($pdl_output->slice(":,(0)"));
	$y_min=-0.5*$stats[0]-0.75*$stats[1];
	$y_max=2*$stats[0]+6*$stats[1];
	($nx,$ny)=$pdl_output->dims();
	$min_wave=$wave_now[0];
	$max_wave=$wave_now[$nx-1];

	pgbegin(0,$dev_plot,1,1);
	pgsfs(1.2);
	pgscf(2);             # Set character font
	pgslw(2);             # Set line width
	pgsch(1.2);           # Set character height
	pgenv($min_wave,$max_wave,$y_min,$y_max,0,0);
	pgsch(0.9);           # Set character height
	pglabel("Wavelength","Flux","$title");
	pgsch(0.5);
	for ($j=0;$j<$ny;$j++) {
	    my $t=$pdl_output->slice(":,($j)");
	    pgsci(1+$j);
	    my @flux=list($t);
	    pgline($nx,\@wave_now,\@flux);
	}
	pgsci(1);
	pgclose;
	pgend;
    }
}

sub plot_test_external {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_model=$_[2];
    my $pdl_model_cont=$_[3];
    my ($n)=$pdl_wave->dims();
    my $i;
    open(OUT,">out_mod_res.fit_spectra");
    my $i;
    for ($i=0;$i<$n;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
	my $f_mod=$pdl_model->at($i);#*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
    if ($plot==1) {
	$call="plot_out_fit_mod.pl out_mod_res.fit_spectra 1/xs ";
	system($call);
    }
}


sub create_single_model() {
    my $pdl_wave=$_[0];
    my $i_now=$_[1];
    my $junk=$_[2];
    my @type=@$junk;
    my $junk=$_[3];
    my @a_c=@$junk;

    my @dims=$pdl_wave->dims();
    my $nx=$dims[0];
    my $i=0;
    my $pdl_out=zeroes($nx);
    my $j=0;

    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	if ($type[$i_now] eq "eline") {
	    my $speed_of_light=299792.458;
	    my $factor=(1+$a_c[$i_now][3]/$speed_of_light+$a_c[$i_now][5]);
	    my $e1=1;
	    my $Y1=1;
	    if ($a_c[$i_now][2]!=0) {
		$e1=exp(-0.5*(($w-$a_c[$i_now][0]*$factor)/$a_c[$i_now][2])**2);
		$Y1=$a_c[$i_now][1]*$e1/($a_c[$i_now][2]*((2*3.1416)**0.5));
	    }
	    set($pdl_out,$i,$Y1);
	}
	if ($type[$i_now] eq "poly1d") {
	    my $ii;
	    my $Yi=0;
	    for ($ii=0;$ii<9;$ii++) {
		$Yi=$Yi+$a_c[$i_now][$ii]*($w)**($ii);
	    }
#	    print "$Yi\n";
	    set($pdl_out,$i,$Yi);
	}
    }
    return $pdl_out;
}

sub print_a_val() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @e_a_print=@$junk;
    my $junk=$_[3];
    my @type=@$junk;
    my $chi_a=$_[4];
    my $ii,$j;
    print "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print "$a_print[$ii][$j] $e_a_print[$ii][$j] ";
	}
	print "\n";
    }
}


sub print_a_final() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $ii,$j;
    print "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print "\n";
    }
}

sub print_a_final_file() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $outfile=$_[4];
    my $ii,$j;
    print "OUT_FILE = $outfile\n";
    open(OUT,">$outfile");
    print OUT "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print OUT "\n";
    }
    close(OUT);
}


sub print_a_final_file_add() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $outfile=$_[4];
    my $ii,$j;
    print "OUT_FILE = $outfile\n";
    open(OUT,">>$outfile");
    print OUT "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print OUT "\n";
    }
    close(OUT);
}




#print_config_file($n_mod,$chi_sq_goal,$d_chi_sq_goal,\@a_final,\@type,\@ia,\@a0,\@a1,\@link,$out_config);

sub print_config_file() {
    my $n_mod=$_[0];
    my $chi_sq_goal=$_[1];
    my $d_chi_sq_goal=$_[2];
    my $junk=$_[3];
    my @a_print=@$junk;
    my $junk=$_[4];
    my @type=@$junk;
    my $junk=$_[5];
    my @ia=@$junk;
    my $junk=$_[6];
    my @a0=@$junk;
    my $junk=$_[7];
    my @a1=@$junk;
    my $junk=$_[8];
    my @link=@$junk;
    my $outfile=$_[9];
    my $ii,$j;
    print "OUT_CONFIG = $outfile\n";
    open(OUT,">$outfile");
    print OUT "0 $n_mod $chi_sq_goal $d_chi_sq_goal\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii]\n";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $ia[$ii][$j] $a0[$ii][$j] $a1[$ii][$j] $link[$ii][$jj]\n";
	}
    }
    close(OUT);
}


sub copy_a() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$ii][$j]=$a_print[$ii][$j];
	}
    }
    return @a_copy;
}

sub copy_a_pdl() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$ii][$j]=$junk->at($j,$ii);
	}
    }
    return @a_copy;
}

sub random_a() {
    my $n_mod=$_[0];
    my $r_scale=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {
	my $pdl_rnd=grandom(10);
	for ($j=0;$j<9;$j++) {	    
		my $rnd=$pdl_rnd->at($j); 
		if (($j!=3)&&($link[$ii][$j]==-1)) {		
		    $a_copy[$ii][$j]=$a_print[$ii][$j]+$r_scale*$rnd*$a_print[$ii][$j];
		} else {
		    $a_copy[$ii][$j]=$a_print[$ii][$j];
		}
	    
	}
    }
    return @a_copy;
}

#	($n_mod_fixed,\@a_fixed,\@a_type_fixed)=add_a_results($n_mod,\@a_final,\@a_type,$n_mod_fixed,\@a_fixed,\@a_type_fixed);    
sub add_a_results_elines() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_final=@$junk;
    my $junk=$_[2];
    my @a_type=@$junk;
    my $n_mod_fixed=$_[3];
    my $junk=$_[4];
    my @a_final_fixed=@$junk;
    my $junk=$_[5];
    my @a_type_fixed=@$junk;
    my $ii,$j;
    my $KK=0;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	if ($a_type[$ii] eq "eline") {
	    $a_type_fixed[$KK+$n_mod_fixed]=$a_type[$ii];	
	    for ($j=0;$j<9;$j++) {
		$a_final_fixed[$KK+$n_mod_fixed][$j]=$a_final[0][$ii][$j];
	    }
	    $KK++;
	}
    }
    $n_mod_fixed=$n_mod_fixed+$KK;
    return ($n_mod_fixed,\@a_final_fixed,\@a_type_fixed);
}

sub add_a_results() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_final=@$junk;
    my $junk=$_[2];
    my @a_type=@$junk;
    my $n_mod_fixed=$_[3];
    my $junk=$_[4];
    my @a_final_fixed=@$junk;
    my $junk=$_[5];
    my @a_type_fixed=@$junk;
    my $ii,$j;
    my $KK=0;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	$a_type_fixed[$ii+$n_mod_fixed]=$a_type[$ii];	
	for ($j=0;$j<9;$j++) {
	    $a_final_fixed[$ii+$n_mod_fixed][$j]=$a_final[0][$ii][$j];
	}
    }
    $n_mod_fixed=$n_mod_fixed+$n_mod;
    return ($n_mod_fixed,\@a_final_fixed,\@a_type_fixed);
}

sub copy_a_results() {
    my $n_mod=$_[0];
    my $kk=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my $junk=$_[3];
    my @a_copy=@$junk;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$kk][$ii][$j]=$a_print[$ii][$j];
	}
    }
    return @a_copy;
}

sub mean_a_results() {
    my $n_mod=$_[0];
    my $nk=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j,$k;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    my @a_tmp;
	    for ($k=0;$k<$nk;$k++) { 
		$a_tmp[$k]=$a_print[$k][$ii][$j];
	    }
	    my $val=mean(@a_tmp);
	    my $e_val=sigma(@a_tmp);
	    $a_copy[0][$ii][$j]=$val;
	    $a_copy[1][$ii][$j]=0.5*$e_val*$ia[$ii][$j];
	}
    }
    return @a_copy;
}



sub mean_a_results_last() {
    my $n_mod=$_[0];
    my $nk=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j,$k;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    my @a_tmp;
	    for ($k=0;$k<$nk;$k++) { 
		$a_tmp[$k]=$a_print[$k][$ii][$j];
	    }
	    my $val=mean(@a_tmp);
	    my $e_val=sigma(@a_tmp);
	    $a_copy[0][$ii][$j]=$a_tmp[$nk-1];
	    $a_copy[1][$ii][$j]=0.5*$e_val*$ia[$ii][$j];
	}
    }
    return @a_copy;
}

