#!/usr/bin/perl
#
#
#

#use PGPLOT;

use Statistics::OLS;
use Math::FFT;
use Math::Stat;
use Math::Spline qw(spline linsearch binsearch);
use Math::Derivative qw(Derivative2);
use Math::Approx;
use Astro::FITS::CFITSIO qw( :longnames :constants );
use PDL;
use PDL::Fit::Polynomial; 
use PDL::Filter::Linear;
use PGPLOT;  # Load PGPLOT module
use PDL::Fit::Gaussian;
use forks;

$sec_ini=print_time();

require("/home/sanchez/sda1/perl/MY/my.pl");




#    srand(localtime);

if ($#ARGV<5) {
    print "USE: kin_cube_elines_rnd.pl CUBE.fits CONFIG MASK_LIST W_MIN W_MAX OUTFILE [N_MC=50] [N_LOOPS=15] [PLOT] [SCALE_INI=0.15] [PREFIX] [MEMO=0/1] [input_vel_map.fits,input_vel_mask_now,FIXED=0,1]\n";
    exit;
}

$input_cube=$ARGV[0];
$config=$ARGV[1];
$mask_list=$ARGV[2];
$w_min=$ARGV[3];
$w_max=$ARGV[4];
$out_fit_spectra=$ARGV[5];
$prefix="map_".$out_fit_spectra;

$n_loops=15;
$n_mc=50;
$def=0;

if ($#ARGV==6) {
    $n_mc=$ARGV[6];
}

if ($#ARGV==7) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
}
$plot=0;

if ($#ARGV==8) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
    $plot=$ARGV[8];
}

$scale_ini=0.15;
if ($#ARGV==9) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
    $plot=$ARGV[8];
    $scale_ini=$ARGV[9];
}


if ($#ARGV==10) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
    $plot=$ARGV[8];
    $scale_ini=$ARGV[9];
    $prefix=$ARGV[10];
}

$memo=0;
if ($#ARGV==11) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
    $plot=$ARGV[8];
    $scale_ini=$ARGV[9];
    $prefix=$ARGV[10];
    $memo=$ARGV[11];
}

$guided=0;
if ($#ARGV==12) {
    $n_mc=$ARGV[6];
    $n_loops=$ARGV[7];
    $plot=$ARGV[8];
    $scale_ini=$ARGV[9];
    $prefix=$ARGV[10];
    $memo=$ARGV[11];
    $JUNK=$ARGV[12];
    $guided=1;
    ($guided_map,$guided_mask,$guided_type)=split(",",$JUNK);
    $gmap=rfits($guided_map);
    $gmask=rfits($guided_mask);    
}



##################################################
# READING INPUTS
##################################################

#
# Reading MASK LIST
#
if ($mask_list eq "none") {
    $nmask=0;
} else {
    open(FH,"<$mask_list");
    while($line=<FH>) {
	chop($line);
	@data=split(" ",$line);
	$start_mask[$nmask]=$data[0];
	$end_mask[$nmask]=$data[1];
	$nmask++;
    }
    close(FH);
}

print "$nmask regions to mask\n";

#
# READING FILE!
#
my $pdl_model;
my $nx,$ny,$nz;
$pdl_input_cube=rfits($input_cube);
$h=$pdl_input_cube->gethdr;
($nx,$ny,$nz)=$pdl_input_cube->dims();
$crpix=$pdl_input_cube->hdr->{"CRPIX3"};
$crval=$pdl_input_cube->hdr->{"CRVAL3"};
$cdelt=$pdl_input_cube->hdr->{"CDELT3"};

$pdl_wave_org=$crval+$cdelt*pdl(0..$nz);

$i0=int(($w_min-$crval)/$cdelt);
$i1=int(($w_max-$crval)/$cdelt);

$pdl_flux_cube=$pdl_input_cube->slice(",,$i0:$i1");
$pdl_wave=$pdl_wave_org->slice("$i0:$i1");
$nm=$i1-$i0+1;
$pdl_masked=ones($nm);

$org=zeroes($nx,$ny,$nm,);
$mod=zeroes($nx,$ny,$nm);
$res=zeroes($nx,$ny,$nm);



for ($i=0;$i<$nz;$i++) {
    $w_test=$pdl_wave_org->at($i);
    $masked[$i]=1;
    for ($j=0;$j<$nmask;$j++) {
	if (($w_test>$start_mask[$j])&&($w_test<$end_mask[$j])) {
	    $masked[$i]=0;	    
	}
    }
}
$pdl_masked_org=pdl(@masked);
$pdl_masked=$pdl_masked_org->slice("$i0:$i1");

#
# Reading CONFIG!
#

open(CNF,"<$config");
$cnf=<CNF>; chop($cnf);
($junk,$n_mod,$chi_goal,$d_chi_goal)=split(" ",$cnf);
for ($ii=0;$ii<$n_mod;$ii++) {
    $cnf=<CNF>; chop($cnf);
    $type[$ii]=$cnf;
    for ($j=0;$j<9;$j++) {
	$cnf=<CNF>; chop($cnf);
	@data=split(" ",$cnf);
	$a[$ii][$j]=$data[0];
	$ia[$ii][$j]=$data[1];
	$ea[$ii][$jj]=0;
	$a0[$ii][$j]=$data[2];
	$a1[$ii][$j]=$data[3];
	$link[$ii][$j]=$data[4];
    }


}
close(CNF);
print "$n_mod models to fit\n";

$pdl_f=zeroes($n_mod,$nx,$ny);
$pdl_ef=zeroes($n_mod,$nx,$ny);
$pdl_v=zeroes($n_mod,$nx,$ny);
$pdl_ev=zeroes($n_mod,$nx,$ny);
$pdl_dw=zeroes($n_mod,$nx,$ny);
$pdl_edw=zeroes($n_mod,$nx,$ny);




my $jy,$ix;
my @a_final_results;
#print "($nx,$ny,$nz)\n";
for ($ix=0;$ix<$nx;$ix++) {
    for ($jy=0;$jy<$ny;$jy++) {
	
    $pdl_flux=$pdl_flux_cube->slice("($ix),($jy),");
    $pdl_flux->inplace->setbadtoval(0); 
    @stats=stats($pdl_flux);    
    $pdl_e_flux=sqrt(abs($pdl_flux+$stats[1]/$nm));
    $pdl_model=zeroes($nm);
    $pdl_model_cont=zeroes($nm);




    $chi_sq_now=5;
    $chi_sq_start=$chi_sq_now;
    my @a_out_now;
    my @a_results;
    my $nnk=0;
    my @a_out_tmp;
    my $i_ter=0;
    @a_out_now=copy_a($n_mod,\@a);
    @ia_now=copy_a($n_mod,\@ia);
    @a0_now=copy_a($n_mod,\@a0);
    @a1_now=copy_a($n_mod,\@a1);

#
# We forced a certain velocity!
#
    if ($guided==1) {
	for ($j=0;$j<$n_mod;$j++) {
	    my $mask=$gmask->at($ix,$jy);
	    if ($link[$j][3]==-1) {
		if (($type[$j] eq "eline")&&($mask==1)) {	    
		    $a_out_now[$j][3]=$gmap->at($ix,$jy);
		    if ($guided_type==0) {
			$ia_now[$j][3]=0;
		    }
		    if ($guided_type==1) {
			$ia_now[$j][3]=1;
			$a0_now[$j][3]=$a_out_now[$j][3]-7;
			$a1_now[$j][3]=$a_out_now[$j][3]+7;
		    }
		    
		}
	    }
	}
    }


    $break=0;

    if (($stats[0]!=$stats[1])||($stats[0]!=0)) {

	srand(localtime);
	while (($i_ter<$n_loops)&&($break==0)) {
	    my $chi_iter=$chi_sq_now;
	    @a_out_now=fit_elines_grad_rnd_fork($pdl_wave,$pdl_flux,$pdl_e_flux,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out_now,\@ia_now,\@a0_now,\@a1_now,\@link,$n_mc,$pdl_masked,$def,$scale_ini,5);
	    if ($chi_sq_now<$chi_iter) {
		@a_out_tmp=copy_a($n_mod,\@a_out_now);
		@a_results=copy_a_results($n_mod,$nnk,\@a_out_now,\@a_results);    
		$nnk++;
		$i_ter++;
	    } else {
		my $rnd_a=random(10);
		@a_out_now=copy_a($n_mod,\@a_out_now);
		$break=1;
	    }
	}
	
	
	if ($plot==1) {
	    plot_test();
	    print "- $jy/$ny ---------------------------------------\n";
	}
	
    } 

    @a_final=mean_a_results($n_mod,$nnk,\@a_results,$ix,$jy,\@type,$pdl_wave);
    if ($plot==1) {
	print_a_final($n_mod,\@a_final,\@type,$chi_sq_now);
    }
    if (($jy==0)&&($ix==0)) {
	print_a_final_file($n_mod,\@a_final,\@type,$chi_sq_now,$out_fit_spectra);
    } else {
	add_a_final_file($n_mod,\@a_final,\@type,$chi_sq_now,$out_fit_spectra);
    }

    my $t=$org->slice("($ix),($jy),:");
    $t .= $pdl_flux;
    my $t=$mod->slice("($ix),($jy),:");
    $t .= $pdl_model;
    my $t=$res->slice("($ix),($jy),:");
    $t .= $pdl_flux-$pdl_model;

    if (($memo==1)&&($chi_sq_now<2*$chi_goal)) {
	@a_out_now=copy_a_final($n_mod,\@a_final);	
    }
    

    }
}

$$h{"CRPIX3"}=1;
$$h{"CRVAL3"}=$crval+$i0*$cdelt;
$$h{"CDELT3"}=$cdelt;
$$h{"NAXIS3"}=$nm;
$org->sethdr($h);
$org->wfits("kin_back_cube_org.fits");
$res->sethdr($h);
$res->wfits("kin_back_cube_res.fits");
$mod->sethdr($h);
$mod->wfits("kin_back_cube_mod.fits");

#
# We create the maps!
#

print "Creating maps...\n";
for ($i=0;$i<$n_mod;$i++) {
    if ($type[$i] ne "poly1d") {
	
	my $sec_f=$pdl_f->slice("($i),,");
	my $sec_ef=$pdl_ef->slice("($i),,");
	my $sec_v=$pdl_v->slice("($i),,");
	my $sec_ev=$pdl_ev->slice("($i),,");
	my $sec_dw=$pdl_dw->slice("($i),,");
	my $sec_edw=$pdl_edw->slice("($i),,");
	if($i<10) {
	    $num="0".$i;
	} else {
	    $num=$i;
    }
	$outfile=$prefix."_".$num.".fits";
	$outfile_f=$prefix."_flux_".$num.".fits";
	$outfile_ef=$prefix."_eflux_".$num.".fits";
	$outfile_w=$prefix."_vel_".$num.".fits";
	$outfile_ew=$prefix."_evel_".$num.".fits";
	$outfile_dw=$prefix."_disp_".$num.".fits";
	$outfile_edw=$prefix."_edisp_".$num.".fits";
	$sec_f->wfits($outfile_f);
	$sec_ef->wfits($outfile_ef);
	$sec_v->wfits($outfile_w);
	$sec_ev->wfits($outfile_ew);
	$sec_dw->wfits($outfile_dw);
	$sec_edw->wfits($outfile_edw);
    } else {
      
	my $sec_f=$pdl_f->slice("($i),,");
	my $sec_ef=$pdl_ef->slice("($i),,");
	$outfile_f=$prefix."_cont.fits";
	$outfile_ef=$prefix."_econt.fits";
	$sec_f->wfits($outfile_f);
	$sec_ef->wfits($outfile_ef);

    }
}



$sec_end=print_time();

$sec_total=$sec_end-$sec_ini;
print "# SECONDS = $sec_total\n";

exit;


sub print_time {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    print "# TIME $sec $min $hour $mday $mon $year $wday $yday $isdst\n";
    print LOG "# TIME $sec $min $hour $mday $mon $year $wday $yday $isdst\n";
    my $sec_now=$min*60+$sec;
    return $sec_now;
}


sub fit_elines_rnd_OLD() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_rnd=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
    my $SCALE=$SCALE_INI;
    my $nx,$ny,$nz;
    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }
#    print "$nx,$ny\n";
#    my $pdl_model=zeroes($nx,$ny);
#    my $pdl_masked=ones($nx);

    my $i;
    #
    #

    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);

    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
#    print "CHI_sq [GUESS]=$chi_sq\n";  
	$chi_sq_ini=$chi_sq;
	@a_out=copy_a($n_mod,\@a);




    #############################
    # TEST OUTPUT!
    #
    open(OUT,">test_fit_elines_rnd.spec");
    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
	my $f_mod=$pdl_model_0->at($i)*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont_0->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6450 6800 &";
    system($call);
#    print "###################################\n";
#    <stdin>;
    #
    # TEST OUTOUT
    ###############################

    my $i,$j,$ii;
    my @a_now;

#    srand(localtime);
	$pdl_rnd=random($n_mod*9*$n_mc);

    for ($ii=0;$ii<$n_mc;$ii++) {
	for ($i=0;$i<$n_mod;$i++) {
	    for ($j=0;$j<9;$j++) {
		$rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
		if ($ia[$i][$j]==1) {
		    if ($link[$i][$j]==-1) {
			$a_now[$i][$j]=$a0[$i][$j]+($a1[$i][$j]-$a0[$i][$j])*$rnd;	
		    } else {
			$k=$link[$i][$j]-1;
			$method=$a1[$i][$j];
			if ($method==0) {
			    $a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
			} else {
			    $a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
			}			    
		    }
		} else {
		    $a_now[$i][$j]=$a[$i][$j];
		}	    
	    }
	}

    # MC test
    $pdl_model_0=zeroes($nx);
    $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);

    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
	if ($chi_sq<$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;


    #############################
    # TEST OUTPUT!
    #
    open(OUT,">test_fit_elines_rnd.spec");
    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
		my $f_mod=$pdl_model_0->at($i)*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont_0->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
#    print "iter= $ii/$n_mc  CHI_sq=$chi_sq / $chi_sq_ini SCALE = $SCALE\n";	    
#	    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6300 7000 ";
    $call="plot_out_fit_mod.pl test_fit_elines_rnd.spec 1/xs 0 2.5e4 6450 6800 &";
    system($call);
    #
    # TEST OUTOUT
	    ###############################


	    my $iii;
	    for ($iii=0;$iii<$n_mod-1;$iii++) {
		if (($type[$iii] eq "eline") and ($link[$iii][1]==-1)) {
		    $a0[$iii][1]=$a_out[$iii][1]/(1+1/$SCALE);
		    $a1[$iii][1]=$a_out[$iii][1]*(1+$SCALE);
		}
	    }


	    $SCALE=0.75*$SCALE;
	    if ($SCALE>0.001) {
		$ii=0;
	    } else {
		$ii=$n_mc;
	    }








	} else {

	#    $SCALE=1.1*$SCALE;

	}
#else {	    
#	    if (($def==1)&&($chi_sq>2*$chi_sq_ini)) {
#	    if ($def==3) {
#
#		}
#	    }
#	}

	   
    }
    $chi_sq_now=$chi_sq_ini;





    return @a_out;
}


sub fit_elines_grad_rnd() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_rnd=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
    my $SCALE=$SCALE_INI;


    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;

    my $pdl_rnd=grandom($n_mod*9*$n_mc);
    my $pdl_rnd_lin=random($n_mod*9*$n_mc);

    # We start the fitting loop!
    my $ii=0;
    while ($ii<$n_mc) {
	my $pdl_chi_now;
	my $chi_sq;
	my $II;
	my @childs;
	my $pid;


	my $junk=fit_elines_rnd_single($pdl_wave,$pdl_flux,$pdl_e_flux,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out,\@ia,\@a0,\@a1,\@link,$n_mc,$pdl_masked,$def,$scale_ini,$pdl_rnd,$pdl_rnd_lin,$ii);
	my @a_now=@$junk;
	$chi_sq=$a_now[$n_mod][0];

#	$chi_sq=$chi_single;

#	print "$ii,$n_mc,$chi_sq,$chi_sq_ini\n";

	
	if ($chi_sq<$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;
	    $pdl_model=$pdl_model_0;
	    $pdl_model_cont=$pdl_model_cont_0;
	} else {
	    if ((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal)) {
		$ii=$n_mc;
	    }
	}

	$ii++;
    }
    $chi_sq_now=$chi_sq_ini;
    return @a_out;
}



sub fit_elines_grad_rnd_fork() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_rnd=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my $NC=$_[16];
    my @a_out;    
    my $SCALE=$SCALE_INI;

    my @a_fork;
    
    my $i; 


    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;

    my $pdl_rnd=grandom($n_mod*9*$n_mc);
    my $pdl_rnd_lin=random($n_mod*9*$n_mc);

    # We start the fitting loop!
    my $ii=0;
    my @a_now_final;
    my @a_now_single;

    while ($ii<$n_mc) {
	my $pdl_chi_now;
	my $chi_sq;
	my $II;
	my @childs;
	my $pid;

# $forks[$i] = async { return sum($i,$sum); };

	for ($II=0;$II<$NC;$II++) {
	    $a_now_single[$II]= async {     my $pdl_rnd=grandom($n_mod*9*$n_mc);
					    my $pdl_rnd_lin=random($n_mod*9*$n_mc);
					    fit_elines_rnd_single($pdl_wave,$pdl_flux,$pdl_e_flux,$n_mod,$chi_goal,$d_chi_goal,\@type,\@a_out,\@ia,\@a0,\@a1,\@link,$n_mc,$pdl_masked,$def,$scale_ini,$pdl_rnd,$pdl_rnd_lin,$ii);
	    }
	}
	

	for ($II=0;$II<$NC;$II++) {
	    my $junk = $a_now_single[$II]->join();
	    @a_now=@$junk;
	    $chi_sq=$a_now[$n_mod][0];
#	    print "$ii,$II $chi_sq\n";
	    if ($chi_sq<$chi_sq_ini) {
		@a_out=copy_a($n_mod,\@a_now);
		$chi_sq_ini=$chi_sq;
		$pdl_model=$pdl_model_0;
		$pdl_model_cont=$pdl_model_cont_0;
	    } else {
		if ((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal)) {
		    $ii=$n_mc;
		}
	    }
	    
	}
	$ii=$ii+$NC;
	$chi_sq_now=$chi_sq_ini;
    }
    return @a_out;
}



sub fit_elines_rnd_single() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_rnd=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
    my $SCALE=$SCALE_INI;
    my $pdl_rnd=$_[16];
    my $pdl_rnd_lin=$_[17];
    my $ii=$_[18];
    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
	$ny=1;
    }

    if ($#dims==1) {
	$nx=$dims[0];
	$ny=$dims[1];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;


    # We start the fitting loop!

    my $pdl_chi_now;
    my $chi_sq;
    my $II;
    my @childs;
    my $pid;


#    print_a_final($n_mod,\@a_out,\@type,$chi_sq_now);

    
	# We change slighly the parameters
    #
    for ($i=0;$i<$n_mod;$i++) {
	for ($j=0;$j<9;$j++) {
	    $rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
	    $rnd_lin=$pdl_rnd_lin->at(($j+(9*$i))*$ii);
	    if ($ia[$i][$j]==1) {
		if ($link[$i][$j]==-1) {
		    if ($j==3) {
			$a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$rnd*($a1[$i][$j]-$a0[$i][$j])/5;
		    } else {
			$a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$a_out[$i][$j]*$rnd;
		    }
		    if ($a_now[$i][$j]<$a0[$i][$j]) {
			$a_now[$i][$j]=$a0[$i][$j];
		    }
		    if ($a_now[$i][$j]>$a1[$i][$j]) {
			$a_now[$i][$j]=$a1[$i][$j];
		    }
		} else {
		    $k=$link[$i][$j]-1;
		    $method=$a1[$i][$j];
		    if ($method==0) {
			$a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
		    } else {
			$a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
		    }			    
		}
	    } else {
		$a_now[$i][$j]=$a_out[$i][$j];
	    }	    
#	    print "$i,$j $a_now[$i][$j]\n";
	}
    }
    
    # MC test
    $pdl_model_0=zeroes($nx);
    $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);
    $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_single=$chi_sq;
    $a_now[$n_mod][0]=$chi_sq;
#    print "@a_now\n";
    #$chi_sq_now=$chi_sq;
    #print "$chi_sq_now\n";

    return \@a_now;
}




sub get_initial_guess() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $i_now=$_[2];
    my $junk=$_[3];
    my @type=@$junk;
    my $junk=$_[4];
    my @a=@$junk;
    my @dims=$pdl_wave->dims();
    my $nx=$dims[0];
    my $i=0;
#    my $pdl_out=zeroes($nx);
    my $j=0;
    my $guess;

    if ($type[$i_now] eq "eline") {
	$guess=$a[$i_now][1];
	$min_delta_wave=1e12;
	for ($i=0;$i<$nx;$i++) {
	    my $w=$pdl_wave->at($i);	    
	    my $speed_of_light=299792.458;
	    my $factor=(1+$a[$i_now][3]/$speed_of_light+$a[$i_now][5]);
	    my $delta_wave=abs($w-$a[$i_now][0]*$factor);
	    if ($min_delta_wave>$delta_wave) {
		$min_delta_wave=$delta_wave;
		$guess=$pdl_flux->at($i);
		$guess=$guess*($a[$i_now][2]*((2*3.1416)**0.5));
	    }
	}
    }

    if ($type[$i_now] eq "poly1d") {
	$guess=$a[$i_now][0];
	@stats=stats($pdl_flux);
	$guess=$stats[2];
    }
    return $guess;
}



sub fit_elines_grad_rnd() {
    my $pdl_wave=$_[0];
    my $pdl_flux=$_[1];
    my $pdl_e_flux=$_[2];
    my $n_mod=$_[3];
    my $chi_goal=$_[4];
    my $d_chi_goal=$_[5];
    my $junk=$_[6];
    my @type=@$junk;
    my $junk=$_[7];
    my @a=@$junk;
    my $junk=$_[8];
    my @ia=@$junk;
    my $junk=$_[9];
    my @a0=@$junk;
    my $junk=$_[10];
    my @a1=@$junk;
    my $junk=$_[11];
    my @link=@$junk;
    my $n_rnd=$_[12];
    my $pdl_masked=$_[13];
    my $def=$_[14];
    my $SCALE_INI=$_[15];
    my @a_out;    
    my $SCALE=$SCALE_INI;

    my $nx,$ny;
    my @dims=$pdl_flux->dims;
    if ($#dims==0) {
	$nx=$dims[0];
    }

    if ($#dims==1) {
	$nx=$dims[0];
    }

    my $i;


    # First Guess
    my $n_points=sum($pdl_masked);
    my $pdl_model_0=zeroes($nx);
    my $pdl_model_cont_0=zeroes($nx);
    for ($i=0;$i<$n_mod;$i++) {
	my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a);
	$pdl_model_0=$pdl_model_0+$pdl_tmp;
    }
    $pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a);
    my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
    my $chi_sq=sum($pdl_chi_now);
    $chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
    $chi_sq_ini=$chi_sq;
    @a_out=copy_a($n_mod,\@a);


    my $i,$j,$ii;
    my @a_now;


    # We start the fitting loop!
    $pdl_rnd=grandom($n_mod*9*$n_mc);
    $pdl_rnd_lin=random($n_mod*9*$n_mc);
    my $ii=0;
    while ($ii<$n_mc) {

	#
	# We change slighly the parameters
	#
	for ($i=0;$i<$n_mod;$i++) {
	    for ($j=0;$j<9;$j++) {
		$rnd=$pdl_rnd->at(($j+(9*$i))*$ii);
		$rnd_lin=$pdl_rnd_lin->at(($j+(9*$i))*$ii);
		if ($ia[$i][$j]==1) {
		    if ($link[$i][$j]==-1) {
			if ($j==3) {
			    $a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$rnd*($a1[$i][$j]-$a0[$i][$j])/5;
			} else {
			    $a_now[$i][$j]=$a_out[$i][$j]+$SCALE*$a_out[$i][$j]*$rnd;
			}
			if ($a_now[$i][$j]<$a0[$i][$j]) {
			    $a_now[$i][$j]=$a0[$i][$j];
			}
			if ($a_now[$i][$j]>$a1[$i][$j]) {
			    $a_now[$i][$j]=$a1[$i][$j];
			}
		    } else {
			$k=$link[$i][$j]-1;
			$method=$a1[$i][$j];
			if ($method==0) {
			    $a_now[$i][$j]=$a_now[$k][$j]+$a0[$i][$j];
			} else {
			    $a_now[$i][$j]=$a_now[$k][$j]*$a0[$i][$j];
			}			    
		    }
		} else {
		    $a_now[$i][$j]=$a_out[$i][$j];
		}	    
	    }
	}

	# MC test
	$pdl_model_0=zeroes($nx);
	$pdl_model_cont_0=zeroes($nx);
	for ($i=0;$i<$n_mod;$i++) {
	    my $pdl_tmp=create_single_model($pdl_wave,$i,\@type,\@a_now);
	    $pdl_model_0=$pdl_model_0+$pdl_tmp;
	}
	$pdl_model_cont_0=create_single_model($pdl_wave,($n_mod-1),\@type,\@a_now);
	my $pdl_chi_now=$pdl_masked*(($pdl_flux-$pdl_model_0)**2)/($pdl_e_flux**2);
	my $chi_sq=sum($pdl_chi_now);
	$chi_sq=($chi_sq/($n_points-$n_mod-1))**0.5;
	if ($chi_sq<$chi_sq_ini) {
	    @a_out=copy_a($n_mod,\@a_now);
	    $chi_sq_ini=$chi_sq;
	    $pdl_model=$pdl_model_0;
	    $pdl_model_cont=$pdl_model_cont_0;
#	    $ii=0;
	} else {
	    if (($ii>2)&&((abs($chi_sq-$chi_sq_ini)<$d_chi_sq_goal)||($chi_sq_ini<$chi_sq_goal))) {
		$ii=$n_mc;
	    }
	}
	$ii++;
    }
    $chi_sq_now=$chi_sq_ini;
    return @a_out;
}

sub plot_test {
    open(OUT,">out_mod_res.fit_spectra");
    my $i;
    my @dim=$pdl_wave->dims();
    my $n=$dim[0];

    for ($i=0;$i<$n;$i++) {
	my $w=$pdl_wave->at($i);
	my $f=$pdl_flux->at($i);
	my $f_mod=$pdl_model->at($i)*($pdl_masked->at($i));
	my $f_res=$f-$f_mod;
	my $f_cont=$pdl_model_cont->at($i);
	print OUT "$w $f $f_mod $f_red $f_cont\n";
    }
    close(OUT);
    if ($plot==1) {
	$call="plot_out_fit_mod.pl out_mod_res.fit_spectra 1/xs &";
	system($call);
    }
   # <stdin>;
}


sub create_single_model() {
    my $pdl_wave=$_[0];
    my $i_now=$_[1];
    my $junk=$_[2];
    my @type=@$junk;
    my $junk=$_[3];
    my @a_c=@$junk;

    my @dims=$pdl_wave->dims();
    my $nx=$dims[0];
    my $i=0;
    my $pdl_out=zeroes($nx);
    my $j=0;

    for ($i=0;$i<$nx;$i++) {
	my $w=$pdl_wave->at($i);
	if ($type[$i_now] eq "eline") {
	    my $speed_of_light=299792.458;
	    my $factor=(1+$a_c[$i_now][3]/$speed_of_light+$a_c[$i_now][5]);
	    my $e1=exp(-0.5*(($w-$a_c[$i_now][0]*$factor)/$a_c[$i_now][2])**2);
	    my $Y1=$a_c[$i_now][1]*$e1/($a_c[$i_now][2]*((2*3.1416)**0.5));
	    set($pdl_out,$i,$Y1);
	}
	if ($type[$i_now] eq "poly1d") {
	    my $ii;
	    my $Yi=0;
	    for ($ii=0;$ii<9;$ii++) {
		$Yi=$Yi+$a_c[$i_now][$ii]*($w)**($ii);
	    }
#	    print "$Yi\n";
	    set($pdl_out,$i,$Yi);
	}
    }
    return $pdl_out;
}

sub print_a_val() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @e_a_print=@$junk;
    my $junk=$_[3];
    my @type=@$junk;
    my $chi_a=$_[4];
    my $ii,$j;
    print "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print "$a_print[$ii][$j] $e_a_print[$ii][$j] ";
	}
	print "\n";
    }
}


sub print_a_final() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $ii,$j;
    print "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print "\n";
    }
}

sub print_a_final_file() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $outfile=$_[4];
    my $ii,$j;
    open(OUT,">$outfile");
    print OUT "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print OUT "\n";
    }
    close(OUT);
}

sub add_a_final_file() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my $junk=$_[2];
    my @type=@$junk;
    my $chi_a=$_[3];
    my $outfile=$_[4];
    my $ii,$j;
    open(OUT,">>$outfile");
    print OUT "$n_mod $chi_a\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii] ";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $a_print[1][$ii][$j] ";
	}
	print OUT "\n";
    }
    close(OUT);
}


#print_config_file($n_mod,$chi_sq_goal,$d_chi_sq_goal,\@a_final,\@type,\@ia,\@a0,\@a1,\@link,$out_config);

sub print_config_file() {
    my $n_mod=$_[0];
    my $chi_sq_goal=$_[1];
    my $d_chi_sq_goal=$_[2];
    my $junk=$_[3];
    my @a_print=@$junk;
    my $junk=$_[4];
    my @type=@$junk;
    my $junk=$_[5];
    my @ia=@$junk;
    my $junk=$_[6];
    my @a0=@$junk;
    my $junk=$_[7];
    my @a1=@$junk;
    my $junk=$_[8];
    my @link=@$junk;
    my $outfile=$_[9];
    my $ii,$j;
    print "OUT_CONFIG = $outfile\n";
    open(OUT,">$outfile");
    print OUT "0 $n_mod $chi_sq_goal $d_chi_sq_goal\n";
    for ($ii=0;$ii<$n_mod;$ii++) {   
	print OUT "$type[$ii]\n";
	for ($j=0;$j<9;$j++) {
	    print OUT "$a_print[0][$ii][$j] $ia[$ii][$j] $a0[$ii][$j] $a1[$ii][$j] $link[$ii][$jj]\n";
	}
    }
    close(OUT);
}


sub copy_a() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$ii][$j]=$a_print[$ii][$j];
	}
    }
    return @a_copy;
}

sub copy_a_final() {
    my $n_mod=$_[0];
    my $junk=$_[1];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$ii][$j]=$a_print[0][$ii][$j];
	}
    }
    return @a_copy;
}

sub random_a() {
    my $n_mod=$_[0];
    my $r_scale=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my @a_copy;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {
	my $pdl_rnd=grandom(10);
	for ($j=0;$j<9;$j++) {	    
		my $rnd=$pdl_rnd->at($j); 
		if (($j!=3)&&($link[$ii][$j]==-1)) {		
		    $a_copy[$ii][$j]=$a_print[$ii][$j]+$r_scale*$rnd*$a_print[$ii][$j];
		} else {
		    $a_copy[$ii][$j]=$a_print[$ii][$j];
		}
	    
	}
    }
    return @a_copy;
}


sub copy_a_results() {
    my $n_mod=$_[0];
    my $kk=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my $junk=$_[3];
    my @a_copy=@$junk;
    my $ii,$j;
    for ($ii=0;$ii<$n_mod;$ii++) {   
	for ($j=0;$j<9;$j++) {
	    $a_copy[$kk][$ii][$j]=$a_print[$ii][$j];
	}
    }
    return @a_copy;
}

sub mean_a_results() {
    my $n_mod=$_[0];
    my $nk=$_[1];
    my $junk=$_[2];
    my @a_print=@$junk;
    my $junk=$_[3];
    my @type=@$junk;
    my $pdl_wave=$_[4];
    my @a_copy;
    my $I=$_[3];
    my $J=$_[4];
    my $ii,$j,$k;
    for ($ii=0;$ii<$n_mod;$ii++) {
	for ($j=0;$j<9;$j++) {
	    my @a_tmp;
	    my $kk=0;
	    for ($k=1;$k<$nk;$k++) { 
		$a_tmp[$kk]=$a_print[$k][$ii][$j];
		$kk++;
	    }
	    my $val=0;
	    if ($kk>3) {
		$val=median(@a_tmp);
	    } else {
		$val=mean(@a_tmp);
	    }
	    my $e_val=sigma(@a_tmp);
	    $a_copy[0][$ii][$j]=$val;
	    $a_copy[1][$ii][$j]=$e_val*$ia[$ii][$j];
	}
	if ($type eq "poly1d") {
	    my $sum=0;
	    my $sum_e=0;
	    my @a_wave=list($pdl_wave);
	    my $jj;
	    for ($jj=0;$jj<$#a_wave+1;$jj++) {
		for ($j=0;$j<9;$j++) {
		    $sum=$sum+$a_copy[0][$ii][$j]*($a_wave[$jj])**$j;
		    $sum_e=$sum_e+$a_copy[1][$ii][$j]*($a_wave[$jj])**$j;
		}
	    }
	    set($pdl_f,($ii,$I,$J),$sum);
	    set($pdl_ef,($ii,$I,$J),$sum_e);
	} else {	    
	    set($pdl_f,($ii,$I,$J),$a_copy[0][$ii][1]);
	    set($pdl_ef,($ii,$I,$J),$a_copy[1][$ii][1]);
	    set($pdl_dw,($ii,$I,$J),2.354*$a_copy[0][$ii][2]);
	    set($pdl_edw,($ii,$I,$J),2.354*$a_copy[1][$ii][2]);
	    set($pdl_v,($ii,$I,$J),$a_copy[0][$ii][3]);
	    set($pdl_ev,($ii,$I,$J),$a_copy[0][$ii][3]);
	}
    }
    return @a_copy;
}


sub int_poly_a {
    my $point=$_[0];
    my $wave=$_[1];
    my $suma=0;
    my @data=@$point;
    my $j,$k;
    for ($j=1;$j<$#data;$j=$j+2) {
	$k=($j-1)/2;
	$suma+=$data[$j]*($wave**$k);
    }
    return $suma;
}
