#!/usr/bin/perl
#
#
# This program find peaks in a 2D fiber based spectral image
#
#

use Statistics::OLS;
use Math::FFT;
use Math::Stat;
use Math::Spline qw(spline linsearch binsearch);
use Math::Derivative qw(Derivative2);

use Math::Approx;


use Astro::FITS::CFITSIO qw( :longnames :constants );
use PDL;

use PDL::Fit::Polynomial; 
use PDL::Filter::Linear;
use PGPLOT;  # Load PGPLOT module
use PDL::Fit::Gaussian;
use PDL::Image2D;
use  PDL::Fit::Linfit;



use PDL::Core;
use PDL::Basic;
use PDL::Exporter;
@ISA    = qw( PDL::Exporter );
use PDL::Options ':Func';
use PDL::Slatec; # For matinv()


$vel_light=299792.458;

require("/home/sanchez/sda2/code/R3D/my.pl");

if ($#ARGV<4) {
    print "USE: map_auto_ssp.pl AUTO.OUT seg_file.fits  PREFIX_OUT WAVE_CEN INST_DIST\n";
    exit;
}

$infile=$ARGV[0];
$seg_file=$ARGV[1];
$prefix=$ARGV[2];
$wave_cen=$ARGV[3];
$inst_disp=$ARGV[4];

$pdl_seg=rfits($seg_file);
($nx,$ny)=$pdl_seg->dims;

$pdl_chi=zeroes($nx,$ny);
$pdl_age=zeroes($nx,$ny);
$pdl_age_mass=zeroes($nx,$ny);
$pdl_e_age=zeroes($nx,$ny);
$pdl_age_lum=zeroes($nx,$ny);
$pdl_e_age_lum=zeroes($nx,$ny);
$pdl_age_lum_M=zeroes($nx,$ny);
$pdl_e_age_lum_M=zeroes($nx,$ny);
$pdl_met_lum=zeroes($nx,$ny);
$pdl_met_mass=zeroes($nx,$ny);
$pdl_met_lumI=zeroes($nx,$ny);
$pdl_met_lumII=zeroes($nx,$ny);
$pdl_met_lumII_mass=zeroes($nx,$ny);
$pdl_e_met_lumII=zeroes($nx,$ny);
$pdl_met_e_lumII_mass=zeroes($nx,$ny);
$pdl_met=zeroes($nx,$ny);
$pdl_e_met=zeroes($nx,$ny);
$pdl_Av=zeroes($nx,$ny);
$pdl_e_Av=zeroes($nx,$ny);
$pdl_disp=zeroes($nx,$ny);
$pdl_vel=zeroes($nx,$ny);
$pdl_e_disp=zeroes($nx,$ny);
$pdl_e_vel=zeroes($nx,$ny);
$pdl_flux=zeroes($nx,$ny);
$pdl_e_flux=zeroes($nx,$ny);
$pdl_ml=zeroes($nx,$ny);
$pdl_mass=zeroes($nx,$ny);

$pdl_disp_km=zeroes($nx,$ny);
$pdl_e_disp_km=zeroes($nx,$ny);

$C_lim=1/100;

print "Reading input files\n";
$nc=-1;
$nc_max=-1;
$coeffs_infile="coeffs_".$infile;
open(FH,"<$coeffs_infile");
while ($line=<FH>) {
    chop($line);
    if ($line !~ "ID") {
	@data=split(" ",$line);
	$i=$data[0];
	if ($i==0) {
	    $nc++;
	    $SUM[$nc]=0;
	}
	$AGE[$i][$nc]=$data[1];
	$MET[$i][$nc]=$data[2];
	$COEFF[$i][$nc]=$data[3];
	$NORM[$i][$nc]=$data[4];
	$ML[$i][$nc]=$data[5];
	$AV[$i][$nc]=$data[6];
	$SUM[$nc]=$SUM[$nc]+$data[3];
#	$lSUM[$nc]=$lSUM[$nc]+log10($data[3]);
	$SUM_MASS[$nc]=$SUM_MASS[$nc]+10**((-1)*$data[5]);

	if ($COEFF[$i][$nc]>$C_lim) {
	    $SUM_now[$nc]=$SUM_now[$nc]+$COEFF[$i][$nc];
	}

#1/$data[5];
#
	if ($nc_max<=$i) {
	    $nc_max=$i+1;
	}
    }
}
close(FH);
#$SUM_now=0;
#print "$i $nc $nc_max $nx,$ny\n";
$pdl_COEFF=zeroes($nx,$ny,$nc_max);
$pdl_NORM=zeroes($nx,$ny,$nc_max);	
$pdl_AV=zeroes($nx,$ny,$nc_max);	
print "DONE\n";

$n=0;
open(FH,"<$infile");
while($line=<FH>) {
    chop($line);
    if ($line !~ "#") {
	@data=split(",",$line);
	$chi[$n]=$data[0];
	$age[$n]=$data[1];
	$e_age[$n]=$data[2];
	$met[$n]=$data[3];
	$e_met[$n]=$data[4];
	$Av[$n]=$data[5];	
	$e_Av[$n]=$data[6];
	$vel[$n]=$data[7]*299792.458;
	$e_vel[$n]=$data[8]*299792.458;
	$disp[$n]=$data[9]/(1+$data[7]);
	$e_disp[$n]=abs($data[10]);
	$flux[$n]=$data[13];
	$e_flux[$n]=$data[14];
	$n++;
    }
}
close(FH);

print "Feeding the arrays\n";
for ($ix=0;$ix<$nx;$ix++) {
    for ($iy=0;$iy<$ny;$iy++) {
	$iseg=$pdl_seg->at($ix,$iy);
	if ($iseg>0) {
	    $iseg=$iseg-1;
	    set($pdl_chi,$ix,$iy,$chi[$iseg]);
	    set($pdl_age,$ix,$iy,$age[$iseg]);
	    set($pdl_met,$ix,$iy,$met[$iseg]);
	    set($pdl_Av,$ix,$iy,$Av[$iseg]);
	    set($pdl_vel,$ix,$iy,$vel[$iseg]);
	    set($pdl_disp,$ix,$iy,$disp[$iseg]);
	    if ($disp[$iseg]>$inst_disp) {
		$disp_km=(sqrt(abs($disp[$iseg]**2-$inst_disp**2))/$wave_cen)*299792.458;
	    } else {
		$disp_km=0;
	    }
	    set($pdl_disp_km,$ix,$iy,$disp_km);
	    set($pdl_flux,$ix,$iy,$flux[$iseg]);
	    set($pdl_e_age,$ix,$iy,$e_age[$iseg]);
	    set($pdl_e_met,$ix,$iy,$e_met[$iseg]);
	    set($pdl_e_Av,$ix,$iy,$e_Av[$iseg]);
	    set($pdl_e_vel,$ix,$iy,$e_vel[$iseg]);
	    set($pdl_e_disp,$ix,$iy,$e_disp[$iseg]);
	    $e_disp_km=($e_disp[$iseg]/$wave_cen)*299792.458;
	    set($pdl_e_disp_km,$ix,$iy,$e_disp_km);
	    set($pdl_e_flux,$ix,$iy,$e_flux[$iseg]);
	    for ($i=0;$i<$nc_max;$i++) {
	#	print "$ix,$iy [$nx,$ny] $i [$nc_max]\n";
		set($pdl_NORM,$ix,$iy,$i,$COEFF[$i][$iseg]);
		$age_now=$AGE[$i][0];
		$met_now=$MET[$i][0];
#		if ($age_now>13.6) {
#		    $age_now=13.6;
#		}

		$age_val=$pdl_age_lum->at($ix,$iy);		
		$age_val=$age_val+$COEFF[$i][$iseg]*(log10($age_now));
		set($pdl_age_lum,$ix,$iy,$age_val);

		if ($SUM_MASS[$i]>0) {
		    $age_val=$pdl_age_mass->at($ix,$iy);		
		    $age_val=$age_val+$COEFF[$i][$iseg]*(log10($age_now))*10**((-1)*$ML[$i][$iseg])/$SUM_MASS[$i];
		    set($pdl_age_mass,$ix,$iy,$age_val);

		    $ml_now=$pdl_ml->at($ix,$iy);
		    $ml_now=$ml_now+$COEFF[$i][$iseg]*$ML[$i][$iseg];
		    set($pdl_ml,$ix,$iy,$ml_now);

		    $mass_now=$pdl_mass->at($ix,$iy);
		    $mass_now=$mass_now+$COEFF[$i][$iseg]*10**($ML[$i][$iseg])*$flux[$iseg]*1e-16;
		    set($pdl_mass,$ix,$iy,$mass_now);
		}

		$met_val=$pdl_met_lum->at($ix,$iy);		
		$met_val=$met_val+$COEFF[$i][$iseg]*$met_now;
		set($pdl_met_lum,$ix,$iy,$met_val);

		$met_val=$pdl_met_lumI->at($ix,$iy);		
		$met_val=$met_val+$NORM[$i][$iseg]*log10($met_now/0.02);
		set($pdl_met_lumI,$ix,$iy,$met_val);

#		if ($SUM[$i]>0) {
#		if (($COEFF[$i][$iseg]>$C_lim)&&($SUM_now[$i]>0)) {
		$met_val=$pdl_met_lumII->at($ix,$iy);	   
		$met_val=$met_val+$COEFF[$i][$iseg]*log10($met_now/0.02);#/$SUM_now[$i];
		set($pdl_met_lumII,$ix,$iy,$met_val);
#		    $SUM_now=$SUM_now+$COEFF[$i][$iseg];
#		}
		
		if ($SUM_MASS[$i]>0) {
		$met_val=$pdl_met_mass->at($ix,$iy);				
		$met_val=$met_val+$COEFF[$i][$iseg]*log10($met_now/0.02)*10**((-1)*$ML[$i][$iseg])/$SUM_MASS[$i];
#*(1/$ML[$i][$iseg])/$SUM_MASS[$i];
#
		set($pdl_met_mass,$ix,$iy,$met_val);
		}

#    $call="write_img_header.pl ".$file_name." AGE ".$AGE[$j][0];
	    }

	    $met_now=$pdl_met->at($ix,$iy);
	    $met_e_now=$pdl_e_met->at($ix,$iy);
	    if ($met_now>0) {
		$val_m=$met_e_now/$met_now;
		set($pdl_e_met_lumII,$ix,$iy,$val_m);
	    }
	    $age_now=$pdl_age_lum->at($ix,$iy);
	    $e_age_now=$pdl_e_age->at($ix,$iy);
	    if ($age_now>0) {
		$e_age_now=$e_age_now/$age_now;
	    }
	    set($pdl_e_age_lum,$ix,$iy,$age_val);
	    $age_val=9+$pdl_age_lum->at($ix,$iy);
	    set($pdl_age_lum,$ix,$iy,$age_val);
	    $age_val=9+$pdl_age_mass->at($ix,$iy);
	    set($pdl_age_lum_M,$ix,$iy,$age_val);


#	    $met_val=$pdl_met_lumI->at($ix,$iy);		
#	    $met_val=10**($met_val);
#	    set($pdl_met_lumI,$ix,$iy,$met_val);

	    $met_val=$pdl_met_lum->at($ix,$iy);		
	    $met_val=log10($met_val/0.02);
	    set($pdl_met_lum,$ix,$iy,$met_val);

	}
    }
}
print "DONE\n";
print "Writting the output files\n";

print "'$chi_name'\n";

$chi_name=$prefix."_chi_ssp.fits";
$pdl_chi->wfits($chi_name);

$age_name=$prefix."_age_ssp.fits";
$pdl_age->wfits($age_name);

$age_name=$prefix."_age_mass_ssp.fits";
$pdl_age_mass->wfits($age_name);

$e_age_name=$prefix."_e_age_ssp.fits";
$pdl_e_age->wfits($e_age_name);

$age_name=$prefix."_log_age_yr_ssp.fits";
$pdl_age_lum->wfits($age_name);

$age_name=$prefix."_e_log_age_yr_ssp.fits";
$pdl_e_age_lum->wfits($age_name);

$age_name=$prefix."_log_age_yr_mass_ssp.fits";
$pdl_age_lum_M->wfits($age_name);


$age_name=$prefix."_e_log_age_yr_ssp.fits";
$pdl_e_age_lum->wfits($age_name);

$met_name=$prefix."_met_ssp.fits";
$pdl_met->wfits($met_name);

#$met_name=$prefix."_met_mass_ssp.fits";
$met_name=$prefix."_met_ZH_mass_ssp.fits";
#$pdl_met_mass=$pdl_met_mass-log10(0.02);
$pdl_met_mass->wfits($met_name);

#$met_name=$prefix."_met_lum_ssp.fits";
#$pdl_met_lum->wfits($met_name);

#$met_name=$prefix."_met_lumI_ssp.fits";
#$pdl_met_lumI->wfits($met_name);

$met_name=$prefix."_met_ZH_ssp.fits";
$pdl_met_lumII->wfits($met_name);

$met_name=$prefix."_e_met_ZH_ssp.fits";
$pdl_e_met_lumII->wfits($met_name);

$Av_name=$prefix."_Av_ssp.fits";
$pdl_Av->wfits($Av_name);

$disp_name=$prefix."_disp_ssp.fits";
$pdl_disp->wfits($disp_name);

$disp_km_name=$prefix."_disp_km_h_ssp.fits";
$pdl_disp_km->wfits($disp_km_name);


$vel_name=$prefix."_vel_ssp.fits";
$pdl_vel->wfits($vel_name);

$flux_name=$prefix."_flux_ssp.fits";
$pdl_flux->wfits($flux_name);


$e_met_name=$prefix."_e_met_ssp.fits";
$pdl_e_met->wfits($e_met_name);

$e_Av_name=$prefix."_e_Av_ssp.fits";
$pdl_e_Av->wfits($e_Av_name);

$e_disp_name=$prefix."_e_disp_ssp.fits";
$pdl_e_disp->wfits($e_disp_name);

$e_disp_km_name=$prefix."_e_disp_km_h_ssp.fits";
$pdl_e_disp_km->wfits($e_disp_km_name);

$e_vel_name=$prefix."_e_vel_ssp.fits";
$pdl_e_vel->wfits($e_vel_name);

$e_flux_name=$prefix."_e_flux_ssp.fits";
$pdl_e_flux->wfits($e_flux_name);

$mass_name=$prefix."_mass_ssp.fits";
$pdl_mass->wfits($mass_name);

$ml_name=$prefix."_ml_ssp.fits";
$pdl_ml->wfits($ml_name);

for ($j=0;$j<$nc_max;$j++) {
    $file_name=$prefix."_NORM_".$j."_ssp.fits";
    $sec=$pdl_NORM->slice(":,:,$j");
    $sec->wfits($file_name);
    $call="write_img_header.pl ".$file_name." AGE ".$AGE[$j][0];
    system($call);
    $call="write_img_header.pl ".$file_name." MET ".$MET[$j][0];
    system($call);

}

exit;





